/* BTP library - Banana Tree Protocol
 * Copyright (C) 1999-2001  The Regents of the University of Michigan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */


#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>

#include <glib.h>
#include <gnet/gnet.h>

#include "btp_debug.h"
#include "util.h"

#include "b_conn_connect.h"
#include "b_conn_handler.h"

#include "btp_proto.h"
#include "btp_node.h"
#include "b_packet.h"



static void	btp_conn_connect   (BConn* conn, gpointer user_data);
static void	btp_conn_ping      (BConn* conn, gpointer user_data);
static void	btp_conn_writeable (BConn* conn, gpointer user_data);
static void	btp_conn_close     (BConn* conn, gpointer user_data);
static void	btp_conn_fail      (BConn* conn, gpointer user_data);


BConnFuncs btp_proto_funcs = 
{
  btp_conn_connect,
  btp_conn_ping,
  btp_conn_writeable,
  btp_conn_close,
  btp_conn_fail		/* TODO: Write a close function ? */
};


static void	add_handlers (BtpNode* node);
static void	remove_handlers (BtpNode* node);

static void	receive_any (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_ok_switch (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_error_switch (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_ucast (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_mcast (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_request_neighbors (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_reply_neighbors (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_join (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_switch (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_leave (BConn* conn, BPacket* packet, gpointer user_data);

static void	schedule_request_neighbors (BtpTree* tree, gboolean soon);
static void	cancel_request_neighbors (BtpTree* tree);
static gboolean	request_neighbors_cb (gpointer data);

static void 	schedule_find_new_parent (BtpTree* tree);
static gboolean find_new_parent_cb (gpointer data);
static BtpNode* find_new_parent (BtpTree* tree);

static gboolean	switch_timeout_cb (gpointer data);
static void	schedule_switch_timeout (BtpTree* tree);
static void	cancel_switch_timeout (BtpTree* tree);

#define	TREE	(node->tree)


/* ************************************************************ */



gboolean
btp_peer_handler (BConn* bconn, gpointer user_data)
{
  BtpTree* tree;
  BtpNode* node;

  g_return_val_if_fail (bconn,  TRUE);
  g_return_val_if_fail (bconn->conn,  TRUE);

  BTPP (1, "btp_peer_handler %s:%u\n", 
	bconn->conn->hostname, bconn->conn->port);

  tree = (BtpTree*) user_data;


  node = btp_tree_find_node (tree, bconn->conn->hostname, bconn->conn->port);
  if (node && node->conn) /* Ignore if already connected */
    return TRUE;
  else if (node)	  /* Save connection - we already know about it */
    node->conn = bconn;
  else			  /* New node - create btp node */
    node = btp_node_new (tree, bconn->conn->hostname, bconn->conn->port, bconn);

  add_handlers (node);

  ++tree->num_connect;

  return FALSE;
}




static void
btp_conn_connect (BConn* conn, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;;

  g_return_if_fail (conn);
  g_return_if_fail (node);
  g_return_if_fail (TREE);

  BTPP (1, "btp_conn_connect\n");

  add_handlers (node);

  /* If this is my parent, send it a JOIN and request the siblings
     request. */
  if (btp_tree_is_parent(node))
    {
      BPacket* join_packet;

      /* Send a JOIN */
      join_packet = b_packet_new_join (node);
      b_conn_send_packet (node->conn, join_packet);
      BTPP (7, "SEND JOIN %s:%d\n", node->hostname, node->port);

      /* Schedule a neighbor request. */
      schedule_request_neighbors (TREE, TRUE);

      /* Hint that we have a new parent. */
      if (TREE->hint_func)
	{
	  (TREE->hint_func)(TREE, node, BTP_HINT_NEW_PARENT, TREE->hint_user_data);
	}
    }
}


static void
btp_conn_ping (BConn* bconn, gpointer user_data)
{
  g_return_if_fail (bconn);
  g_return_if_fail (bconn->conn);

  BTPP (5, "PING: %s:%d is %d (%u)\n", 
	bconn->conn->hostname, bconn->conn->port, 
	bconn->distance, millitime());
  /* do nothing for now */
}


static void
btp_conn_writeable (BConn* conn, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;;

  BTPP (1, "btp_conn_writeable\n");

  g_return_if_fail (conn);
  g_return_if_fail (node);
  g_return_if_fail (TREE);

  if (TREE->hint_func)
    {
      (TREE->hint_func)(TREE, node, BTP_HINT_UCAST_WRITE, TREE->hint_user_data);
    }
}


static void
btp_conn_close (BConn* conn, gpointer user_data)
{
  /* Closing and failing are basically the same things. */
  btp_conn_fail (conn, user_data);
}


static void
btp_conn_fail (BConn* conn, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (conn);
  g_return_if_fail (node);
  g_return_if_fail (!TREE || (TREE->me != node));

  BTPP (1, "btp_conn_fail\n");
  BTPP (6, "FAIL: %s:%d\n", node->hostname, node->port);

  remove_handlers (node);

  /* Delete the node if we're deleting it. */
  if (node->is_deleting)
    {
      BTPP (0, "btp_conn_fail: delete\n");

      b_conn_delete (conn);
      memset (node, 0, sizeof (*node));
      g_free (node);
      return;
    }

  /* Notify the upper layer.  Note that we don't notify again if the
     node comes back up. */
  if (TREE->hint_func)
    {
      BTPP (0, "btp_conn_fail: hint\n");

      (TREE->hint_func)(TREE, node, BTP_HINT_NODE_DOWN, TREE->hint_user_data);
    }

  /* If it's our child, delete it */
  if (btp_tree_is_child(node))
    {
      BTPP (0, "btp_conn_fail: is_child\n");

      btp_tree_remove_child (node);
      btp_tree_remove_node (node);
      btp_node_delete (node);
      return;
    }

  /* If it's our parent and the root, do an error upcall */
  else if (btp_tree_is_parent(node) && btp_tree_is_root(node))
    {
      /* TODO: Try reconnecting a few times */

      BTPP (0, "btp_conn_fail: is_parent and root\n");

      /* Do an error upcall */
      if (TREE->error_func)
	{
	  (TREE->error_func)(TREE, TREE->error_user_data);
	  return;
	}
    }

  /* If it's our parent and _not_ the root, disconnect from my parent
     and connect to the root */
  else if (btp_tree_is_parent(node) && !btp_tree_is_root(node))
    {
      /* Set our parent to root */
      TREE->parent = TREE->root;

      BTPP (0, "btp_conn_fail: is_parent, not root\n");

      /* Connect to the root if unconnected. */
      if (b_conn_is_unconnected(TREE->parent->conn))
	b_conn_connect(TREE->parent->conn);

      /* Otherwise, send a JOIN */
      else
	{
	  BPacket* join_packet;

	  /* Send a join */
	  join_packet = b_packet_new_join (TREE->parent);
	  b_conn_send_packet (TREE->parent->conn, join_packet);
	  BTPP (7, "SEND JOIN %s:%d\n", TREE->parent->hostname, TREE->parent->port);

	  /* Schedule a neighbors request. */
	  schedule_request_neighbors (TREE, TRUE);

	  /* Hint that we have a new parent. */
	  if (TREE->hint_func)
	    {
	      (TREE->hint_func)(TREE, TREE->parent, 
				BTP_HINT_NEW_PARENT, TREE->hint_user_data);
	    }
	}
    }

  /* If it's the root and not our parent, let it stay down */
  else if (btp_tree_is_root(node) && !btp_tree_is_parent(node))
    {
      /* Do nothing */;
      BTPP (0, "btp_conn_fail: is root, not parent\n");

    }


  /* If it's our sibling, delete it.  Reset the new_parent stuff if
     it's also our new_parent. */
  else if (btp_tree_is_sibling(node))
    {
      btp_tree_remove_sibling (node);

      BTPP (0, "btp_conn_fail: is sibling\n");

      /* If the node is new_parent, then remove the SWITCH from its
	 queue and schedule a new parent search. */
      if (btp_tree_is_new_parent(node))
	{
	  BTPP (0, "btp_conn_fail: is new parent\n");

	  TREE->new_parent = NULL;

	  /* Cancel the SWITCH timeout */
	  cancel_switch_timeout (TREE);

	  /* Remove the SWITCH packet from the node's out queue */
	  b_conn_remove_packets (node->conn, B_PACKET_TYPE_SWITCH);

	  /* Request neighbors again from parent */
	  schedule_request_neighbors (TREE, TRUE);

	  /* (How do we ensure we don't connect to the node again -
             maybe we should keep it around?) */
	}
      
      btp_tree_remove_node (node);
      btp_node_delete (node);
    }

  /* Otherwise, it's completely unrelated, so delete it */
  else
    {
      BTPP (0, "btp_conn_fail: is no one\n");

      btp_tree_remove_node (node);
      btp_node_delete (node);
    }
}


static void
add_handlers (BtpNode* node)
{
  BConn* conn;

  g_return_if_fail (node);
  g_return_if_fail (node->conn);
  g_return_if_fail (TREE);

  BTPP (1, "add_handlers\n");

  conn = node->conn;

  b_conn_handler_add (conn, B_PACKET_TYPE_OK, B_PACKET_OK_SUBTYPE_SWITCH,
		      receive_ok_switch, node);
  b_conn_handler_add (conn, B_PACKET_TYPE_ERROR, B_PACKET_ERROR_SUBTYPE_SWITCH,
		      receive_error_switch, node);

  b_conn_handler_add (conn, B_PACKET_TYPE_UCAST, 0, receive_ucast, node);
  b_conn_handler_add (conn, B_PACKET_TYPE_MCAST, 0, receive_mcast, node);

  b_conn_handler_add (conn, B_PACKET_TYPE_INFO_REQUEST, 
		      B_PACKET_INFO_SUBTYPE_NEIGHBORS,
		      receive_request_neighbors, node);
  b_conn_handler_add (conn, B_PACKET_TYPE_INFO_REPLY, 
		      B_PACKET_INFO_SUBTYPE_NEIGHBORS,
		      receive_reply_neighbors, node);

  b_conn_handler_add (conn, B_PACKET_TYPE_JOIN, 0, receive_join, node);
  b_conn_handler_add (conn, B_PACKET_TYPE_SWITCH, -1, receive_switch, node);
  b_conn_handler_add (conn, B_PACKET_TYPE_LEAVE, 0, receive_leave, node);

  /* Start handling any packet if there's a packet func */
  if (TREE->packet_func)
    b_conn_handler_add (conn, -1, -1, receive_any, node);
}


static void
remove_handlers (BtpNode* node)
{
  BConn* conn;

  g_return_if_fail (node);
  g_return_if_fail (node->conn);
  g_return_if_fail (TREE);

  BTPP (1, "remove_handlers\n");

  conn = node->conn;

  b_conn_handler_remove (conn, B_PACKET_TYPE_OK, B_PACKET_OK_SUBTYPE_SWITCH, node);
  b_conn_handler_remove (conn, B_PACKET_TYPE_ERROR, B_PACKET_ERROR_SUBTYPE_SWITCH, node);

  b_conn_handler_remove (conn, B_PACKET_TYPE_INFO_REQUEST, 
			 B_PACKET_INFO_SUBTYPE_NEIGHBORS, node);
  b_conn_handler_remove (conn, B_PACKET_TYPE_INFO_REPLY, 
			 B_PACKET_INFO_SUBTYPE_NEIGHBORS, node);

  b_conn_handler_remove (conn, B_PACKET_TYPE_JOIN, 0, node);
  b_conn_handler_remove (conn, B_PACKET_TYPE_SWITCH, 0, node);
  b_conn_handler_remove (conn, B_PACKET_TYPE_LEAVE, 0, node);


  /* Start handling any packet if there's a packet func */
  if (TREE->packet_func)  b_conn_handler_remove (conn, -1, -1, node);

}




/* ************************************************************ */


static void
receive_any (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (packet);
  g_return_if_fail (node);
  g_return_if_fail (TREE);

  BTPP (1, "receive_any\n");

  if (TREE->packet_func)
    TREE->packet_func (TREE, node, packet, TREE->packet_user_data);
}



static void
receive_ok_switch (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  BPacket* new_packet;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_OK);
  g_return_if_fail (packet->subtype == B_PACKET_OK_SUBTYPE_SWITCH);
  g_return_if_fail (TREE);

  BTPP (1, "receive_ok_switch\n");

  ++TREE->num_ok;

  /* Cancel the switch timeout */
  cancel_switch_timeout (TREE);

  /* ******************** */
  /* Checks		*/

  /* We should have a new parent.  It's possible that we btp_fail'ed
     the node and it recovered and sent us this OK. */
  if (TREE->new_parent == NULL)
    return;

  /* This node should be our new parent */
  if (TREE->new_parent != node)
    return;

  /* Don't want to switch to my current parent */
  if (btp_tree_is_parent(node))
    {
      g_warning ("Attempt to switch parents to current parent\n");
      return;
    }

  /* This node should be my sibling or grandparent. (Sanity check -
     this should be correct internally.) */
  if (!btp_tree_is_sibling (node) && !btp_tree_is_gparent (node))
    {
      g_warning ("Received OK from odd relation\n");
      return;
    }


  /* ******************** */
  /* Action		*/

  /* Send a LEAVE to my old parent.  The parent will probably
     close the connection once it receives the LEAVE. */
  new_packet = b_packet_new_leave (TREE->parent);
  b_conn_send_packet (TREE->parent->conn, new_packet);
  BTPP (7, "SEND LEAVE %s:%d\n", TREE->parent->hostname, TREE->parent->port);

  /* Delete my parent now (if it isn't the root).  Note this will
     actually send a BYE to the parent and then delete it. */
  if (!btp_tree_is_root(TREE->parent))
    {
      btp_tree_remove_node (TREE->parent);
      btp_node_delete (TREE->parent);
    }

  BTPP (0, "switch from %s:%d to %s:%d\n",
	TREE->parent->hostname, TREE->parent->port,
	TREE->new_parent->hostname, TREE->new_parent->port);

  /* Switch to this parent */
  btp_tree_remove_sibling (TREE->new_parent);
  TREE->parent = TREE->new_parent;
  TREE->new_parent = NULL;

  /* Delete all siblings */
  while (TREE->siblings)
    {
      BtpNode* sibling = (BtpNode*) TREE->siblings->data;

      g_return_if_fail (TREE->me != sibling);
      g_return_if_fail (TREE->parent != sibling);

      TREE->siblings = g_slist_remove_link (TREE->siblings, TREE->siblings);

      btp_tree_remove_node (sibling);
      btp_node_delete (sibling);
    }

  /* Reset sibling list */
  g_slist_free (TREE->siblings);
  TREE->siblings = NULL;

  /* Get a new list of siblings */
  schedule_request_neighbors (TREE, FALSE);

  /* Give the upper layer a hint that we have a new parent */
  if (TREE->hint_func)
    {
      (TREE->hint_func)(TREE, TREE->parent, BTP_HINT_NEW_PARENT, TREE->hint_user_data);
    }
}



static void
receive_error_switch (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_ERROR);
  g_return_if_fail (packet->subtype == B_PACKET_ERROR_SUBTYPE_SWITCH);
  g_return_if_fail (TREE);

  BTPP (1, "receive_error_switch\n");

  ++TREE->num_error;

  /* Cancel the switch timeout */
  cancel_switch_timeout (TREE);

  /* ******************** */
  /* Checks		*/

  /* We should have a new parent.  It's possible that we
     btp_fail'ed the node and it recovered and sent us this OK.  */
  if (TREE->new_parent == NULL)
    {
      g_warning ("Received weird switch ERROR (I have no new parent)\n");
      return;
    }

  /* ******************** */
  /* Action		*/

  /* Note that this node is a bad parent so that we don't try to
     use it again (we might want to relax this later depending
     on why the switch failed). */
  TREE->new_parent->is_bad_parent = TRUE;

  /* Reset new_parent. */
  TREE->new_parent = NULL;

  /* Request neighbors again.  Presumably we acted on bad information.
     This will start the new parent search process again. */
  schedule_request_neighbors (TREE, TRUE);
}



static void
receive_ucast (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  
  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_UCAST);
  g_return_if_fail (TREE);

  BTPP (1, "receive_ucast\n");

  if (TREE->ucast_read_func)
    {
      TREE->ucast_read_func (TREE, node, packet->data, g_ntohs(packet->length),
			     TREE->ucast_read_user_data);
    }
}


static void
receive_mcast (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  
  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_MCAST);
  g_return_if_fail (TREE);

  BTPP (1, "receive_mcast\n");

  /* Warn if not from relative */
  if (!btp_tree_is_parent(node) && !btp_tree_is_child(node))
    {
      g_warning ("Received mcast packet from non-neighbor.\n");
      return;
    }

  /* Pass up */
  if (TREE->mcast_read_func)
    {
      TREE->mcast_read_func (TREE, node, packet->data, g_ntohs(packet->length),
			     TREE->mcast_read_user_data);
    }
}


static void
receive_request_neighbors (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  BPacket* new_packet;
  
  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_INFO_REQUEST);
  g_return_if_fail (packet->subtype == B_PACKET_INFO_SUBTYPE_NEIGHBORS);
  g_return_if_fail (TREE);

  BTPP (1, "receive_request_neighbors\n");

  new_packet = 
    b_packet_new_info_reply_neighbors (node, TREE->parent, TREE->children);
  b_conn_send_packet (node->conn, new_packet);
  BTPP (7, "SEND INFOREP-NEIGHBORS %s:%d\n", node->hostname, node->port);
}



static void
receive_reply_neighbors (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  BAddress* parent;
  GSList* children;
  GSList* i;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_INFO_REPLY);
  g_return_if_fail (packet->subtype == B_PACKET_INFO_SUBTYPE_NEIGHBORS);
  g_return_if_fail (TREE);

  BTPP (1, "receive_reply_neighbors\n");

  /* Ignore if from a node that isn't my parent or if I'm passive */
  if (node != TREE->parent || TREE->passive)
    return;

  /* Read the neighbors addresses */
  if (b_packet_parse_info_reply_neighbors (packet, &parent, &children))
    {
      g_warning ("Received bad neighbors reply\n");
      schedule_request_neighbors (TREE, FALSE);
      return;
    }

  /* Save gparent */
  TREE->gparent = NULL;
  if (parent && parent->port)
    {
      BtpNode* gparent;

      gparent = btp_tree_find_node (TREE, parent->hostname, parent->port);
      if (gparent == NULL)
	gparent = btp_node_new (TREE, parent->hostname, parent->port, NULL);

      TREE->gparent = gparent;
    }

  /* Save siblings */
  for (i = children; i != NULL; i = i->next)
    {
      BtpNode* sibling;
      BAddress* addr;

      /* Get the next address */
      addr = (BAddress*) i->data;

      /* Ignore if port is 0 */
      if (addr->port == 0)
	{
	  g_warning ("Received child with port of 0\n");
	  continue;
	}

      /* Check if I already have this node */
      sibling = btp_tree_find_node (TREE, addr->hostname, addr->port);
      if (sibling == NULL)
	{
	  sibling = btp_node_new (TREE, addr->hostname, addr->port, NULL);
	  btp_tree_add_sibling (sibling);
	}

      /* Add sibling to sibling list, if not already there.  Don't add
	 if it's me, a parent, sibling, or child. */
      else if (!btp_tree_is_sibling(sibling) &&
	       sibling != TREE->me &&
	       !btp_tree_is_parent(sibling) &&
	       !btp_tree_is_child(sibling))
	{
	  btp_tree_add_sibling (sibling);
	}

      b_address_delete (addr);
    }
  g_slist_free (children);


  /* Connect to all relations to get the ping distance */
  for (i = TREE->siblings; i != NULL; i = i->next)
    {
      BtpNode* sibling = (BtpNode*) i->data;
      b_conn_connect (sibling->conn);
    }
  if (TREE->gparent)
    b_conn_connect (TREE->gparent->conn);

  /* Set a timer to later find a new parent based on ping times */
  schedule_find_new_parent (TREE);
}




static void
receive_join (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_JOIN);
  g_return_if_fail (TREE);

  BTPP (1, "receive_join\n");
  BTPP (7, "RECV JOIN %s:%d\n", node->hostname, node->port);

  ++TREE->num_join;

  /* ******************** */
  /* Checks		*/

  g_return_if_fail (!TREE->passive);		/* Don't allow a JOIN if passive. */
  g_return_if_fail (!btp_tree_is_me(node));  	/* Don't join myself.  */
  g_return_if_fail (!btp_tree_is_root(node));	/* Don't allow root to join. */


  /* ******************** */
  /* Action		*/

  /* If the node is already a child, 'ignore' the JOIN.  This
     might happen if the child resets and rejoins. */
  if (btp_tree_is_child(node))
    return;

  /* If the node is my parent, ignore it.  In practice, this
     shouldn't happen.  It could happen if the parent resets, is
     confused, and rejoins.  However, by the time this happen, I
     should have a new parent. */
  if (btp_tree_is_parent(node))
    return;

  /* Remove it from the sibling list.  This could happen if the
     node resets then sends us a JOIN for some reason. */
  if (btp_tree_is_sibling (node))
    btp_tree_remove_sibling (node);

  /* Add the node to my list of children */
  btp_tree_add_child (node);

  /* Give the upper level a hint */
  if (TREE->hint_func)
    {
      (TREE->hint_func)(TREE, node, BTP_HINT_NODE_JOIN, TREE->hint_user_data);
    }
}



static void
receive_switch (BConn* conn, BPacket* pkt, gpointer user_data)
{
  BtpNode*  node = (BtpNode*) user_data;
  BAddress* parent = NULL;
  gboolean  switch_ok = FALSE;

  g_return_if_fail (node);
  g_return_if_fail (pkt);
  g_return_if_fail (pkt->type == B_PACKET_TYPE_SWITCH);
  g_return_if_fail (TREE);

  BTPP (1, "receive_switch\n");

  ++TREE->num_switch;

  /* ******************** */
  /* Checks		*/

  g_return_if_fail (!TREE->passive);		/* Don't allow a SWITCH if passive */
  g_return_if_fail (!btp_tree_is_me(node));	/* Don't switch to myself */
  g_return_if_fail (!btp_tree_is_parent(node));	/* Don't allow parent to switch */
  g_return_if_fail (!btp_tree_is_root(node));   /* Don't allow root to switch */


  /* ******************** */
  /* Action		*/

  BTPP (7, "RECV SWITCH %s:%d\n", node->hostname, node->port);

  /* If the node is already my child, send an OK. */
  if (btp_tree_is_child(node))
    {
      switch_ok = TRUE;
      goto switch_done;
    }

  /* Can't switch if I'm also trying to switch. */
  if (TREE->new_parent != NULL)
    {
      BTPP (7, "RECV SWITCH ALREADY SWITCHING\n");
      goto switch_done;
    }

  /* Read the parent address.  This is the switching node's parent. */
  if (b_packet_parse_switch(pkt, &parent))
    {
      g_warning ("Received bad switch\n");
      return;
    }

  /* Check relations */
  if (pkt->subtype == B_PACKET_SWITCH_SUBTYPE_SIBLING)
    {
      /* Sibling switches: Can't switch if we have different parents.
	 Otherwise, this might cause a loop. */
      if (parent->port != TREE->parent->port ||
	  strcmp(parent->hostname, TREE->parent->hostname) != 0)
	{
	  goto switch_done;
	}

      /* Ok: Remove from sibling list */
      btp_tree_remove_sibling (node);

      BTPP (7, "RECV SWITCH OK SIBLING\n");
    }
  else if (pkt->subtype == B_PACKET_SWITCH_SUBTYPE_GPARENT)
    {
      BtpNode* kid_node;

      /* Grandparent switches: Can't switch unless node's parent is my
	 child.  Otherwise, this might cause a loop. */
      kid_node = btp_tree_find_node (TREE, parent->hostname, parent->port);
      if (!kid_node || !btp_tree_is_child (kid_node))
	{
	  BTPP (7, "RECV SWITCH ERROR GPARENT: PARENT MISMATCH %s:%d\n",
		parent->hostname, parent->port);
	  goto switch_done;
	}

      BTPP (7, "RECV SWITCH OK GPARENT\n");
    }
  else 
    goto switch_done;

  /* Ok to switch.  Add to my children list. */
  btp_tree_add_child (node);
  switch_ok = TRUE;

 switch_done:

  if (parent)
    {
      g_free(parent->hostname);
      g_free(parent);
    }

  /* Send an OK or error message */
  if (switch_ok)
    {
      BPacket* ok_pkt;

      ok_pkt = b_packet_new_ok (node->conn, B_PACKET_OK_SUBTYPE_SWITCH);
      b_conn_send_packet (node->conn, ok_pkt);
      BTPP (7, "SEND OK-SWITCH %s:%d\n", node->hostname, node->port);
    }
  else
    {
      BPacket* error_pkt;

      ++TREE->num_switch_bad;

      error_pkt = b_packet_new_error (node->conn, B_PACKET_ERROR_SUBTYPE_SWITCH);
      b_conn_send_packet (node->conn, error_pkt);
      BTPP (7, "SEND ERROR-SWITCH %s:%d\n", node->hostname, node->port);
    }
}



static void
receive_leave (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_LEAVE);
  g_return_if_fail (TREE);
  g_return_if_fail (node != TREE->me);

  BTPP (1, "receive_leave\n");
  BTPP (7, "RECV LEAVE %s:%d\n", node->hostname, node->port);

  /* Remove the child */
  if (btp_tree_is_child (node))
    btp_tree_remove_child (node);

  /* Delete the node via btp_conn_fail */
  btp_conn_fail (node->conn, node);
}




/* ************************************************************ */


static void
schedule_request_neighbors (BtpTree* tree, gboolean soon)
{
  guint time;
  guint range;

  g_return_if_fail (tree);

  BTPP (1, "schedule_request_neighbors\n");

  if (tree->request_neighbors_timeout != 0)
    return;

  if (soon)
    {
      time =  REQUEST_NEIGHBORS_TIME;
      range = REQUEST_NEIGHBORS_RANGE;
    }
  else
    {
      time =  NEXT_REQUEST_NEIGHBORS_TIME;
      range = NEXT_REQUEST_NEIGHBORS_RANGE;
    }

  tree->request_neighbors_timeout = 
    g_timeout_add (time + rand() % range, request_neighbors_cb, tree);
}



static void
cancel_request_neighbors (BtpTree* tree)
{
  g_return_if_fail (tree);

  BTPP (1, "cancel_request_neighbor\n");

  if (tree->request_neighbors_timeout)
    {
      g_assert (g_source_remove(tree->request_neighbors_timeout));
      tree->request_neighbors_timeout = 0;
    }
}



static gboolean
request_neighbors_cb (gpointer data)
{
  BtpTree* tree = (BtpTree*) data;
  BPacket* packet;

  BTPP (1, "request_neighbors_cb\n");

  g_return_val_if_fail (tree, FALSE);
  g_return_val_if_fail (tree->parent, FALSE);

  tree->request_neighbors_timeout = 0; 

  packet = b_packet_new_info_request_neighbors (tree->parent);
  b_conn_send_packet (tree->parent->conn, packet);
  BTPP (7, "SEND INFOREQ-NEIGHBORS %s:%d\n", tree->parent->hostname, tree->parent->port);

  return FALSE;
}




/* ************************************************************ */


static void
schedule_find_new_parent (BtpTree* tree)
{
  g_return_if_fail (tree);

  BTPP (1, "schedule_find_new_parent\n");

  if (!tree->find_new_parent_timeout)
    {
      tree->find_new_parent_timeout =
	g_timeout_add (FIND_NEW_PARENT_TIME + rand() % FIND_NEW_PARENT_RANGE, 
		       find_new_parent_cb, tree);
    }
}


static gboolean
find_new_parent_cb (gpointer data)
{
  BtpTree* tree = (BtpTree*) data;
  BtpNode* new_parent;

  BTPP (1, "find_new_parent_cb\n");
  BTPP (7, "FIND NEW PARENT\n");

  g_return_val_if_fail (tree, FALSE);

  /* I should not be root and should have a parent.  TODO: Is is
     possible that we can reach this point and not have a parent? */
  g_return_val_if_fail (tree->root != tree->me, FALSE);
  g_return_val_if_fail (tree->parent != NULL, FALSE);

  /* We should be in switching mode */
  g_return_val_if_fail (tree->find_new_parent_timeout != 0, FALSE);
  tree->find_new_parent_timeout = 0;

  /* We should not have a switching candidate or be in the process of
     switching. */
  g_return_val_if_fail (tree->new_parent == NULL, FALSE);
  g_return_val_if_fail (tree->switch_timeout == 0, FALSE);

  /* Figure out who the new parent should be */
  new_parent = find_new_parent (tree);

  /* Switch if we found a new parent */
  if (new_parent)
    {
      BPacket* switch_packet;
      guint8 subtype;

      BTPP (1, "find_new_parent_cb: Switching parent\n");

      /* Cancel any request neighbor timeout */
      cancel_request_neighbors (tree);

      tree->new_parent = new_parent;
      /* We will remove the parent from the list of siblings in the
	 OK/ERROR code. */

      if (tree->gparent == new_parent)
	{
	  subtype = B_PACKET_SWITCH_SUBTYPE_GPARENT;
	  BTPP (7, "SEND SWITCH GPARENT ");
	}
      else
	{
	  subtype = B_PACKET_SWITCH_SUBTYPE_SIBLING;
	  BTPP (7, "SEND SWITCH SIBLING ");
	}
      BTPP (7, "%s:%d (%d)\n", new_parent->hostname, new_parent->port, new_parent->conn->distance);

      switch_packet = b_packet_new_switch (new_parent, tree->parent, subtype);
      b_conn_send_packet (new_parent->conn, switch_packet);

      schedule_switch_timeout (tree);
    }

  else
    {
      BTPP (0, "find_new_parent_cb: Did not find a better parent\n");

      /* Wait a while, request neighbors, and try again. */
      schedule_request_neighbors(tree, FALSE);
    }

  BTPP (0, "find_new_parent_cb DONE\n");

  return FALSE;
}



static BtpNode*
find_new_parent (BtpTree* tree)
{
  BtpNode* parent;
  BtpNode* gparent;
  BtpNode* closest;
  guint    distance;
  GSList*  si;

  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (tree->parent != NULL, NULL);

  BTPP (1, "find_new_parent\n");

  parent  = tree->parent;
  gparent = tree->gparent;

  /* Closest is our parent (unless we don't know the distance or the
     parent has missed more than one pong) */
  if (parent->conn->distance != 0)
    {
      closest = parent;
      distance = parent->conn->distance;
    }
  else
    {
      closest = NULL;
      distance = G_MAXUINT;
    }

  /* Find the closest sibling */
  for (si = tree->siblings; si != NULL; si = si->next)
    {
      BtpNode* sibling = (BtpNode*) si->data;

      BTPP (10, "sibling %d\n", sibling->conn->distance);

      /* Skip siblings that we don't have a distance for or are known
         bad parents. */
      if (sibling->conn->distance == 0 || sibling->is_bad_parent)
	continue;

      /* If the sibling is closer, then save it */
      if (sibling->conn->distance <= distance)
	{
	  closest  = sibling;
	  distance = sibling->conn->distance;
	}
    }

  /* Check grandparent */
  if (gparent && !(gparent->conn->distance == 0 || gparent->is_bad_parent) &&
      gparent->conn->distance <= distance)
    {
      BTPP (10, "gparent %d\n", gparent->conn->distance);

      closest  = gparent;
      distance = gparent->conn->distance;
    }

  BTPP (10, "closest %p,%d\n", closest, distance);

  /* Return NULL if we didn't find one or didn't find one better than
     our current parent */
  if (closest == NULL || closest == parent)
    return NULL;

  /* Check if aggressively switching (for debugging purposes) */
  if (btp_debug_flags & (1 << 10))
    {
      BTPP (10, "AGGRESSIVE SWITCH\n");
      return closest;
    }

  /* Heuristic: If the closest is a sibling, less than 10 ms away, and
     parent some children, then definately switch to it. */
  if (btp_tree_is_sibling (closest) && 
      closest->conn->distance <= 10 &&
      btp_tree_parent_degree (tree) > MAX_DEGREE_CLOSE)
    return closest;

  /* Otherwise, only switch if it's better than 25% closer _or_ the
     parent has too large a degree.  This means that a node won't
     switch if a sibling is about as close as the parent _and_ the
     parent doesn't have many children.  */
  if (((double) closest->conn->distance <= 
       (.75 * (double) parent->conn->distance)) ||
      (btp_tree_parent_degree (tree) > MAX_DEGREE))
    return closest;

  return NULL;
}



/* ************************************************************ */



static void
schedule_switch_timeout (BtpTree* tree)
{
  g_return_if_fail (tree != NULL);

  BTPP (1, "schedule_switch_timeout\n");

  if (!tree->switch_timeout)
    {
      tree->switch_timeout = 
	g_timeout_add (SWITCH_TIMEOUT_TIME + rand() % SWITCH_TIMEOUT_RANGE, 
		       switch_timeout_cb, tree);
    }
}


static void
cancel_switch_timeout (BtpTree* tree)
{
  g_return_if_fail (tree != NULL);

  BTPP (1, "cancel_schedule_switch_timeout\n");

  if (tree->switch_timeout)
    {
      g_assert (g_source_remove(tree->switch_timeout));
      tree->switch_timeout = 0;
    }
}


static gboolean
switch_timeout_cb (gpointer data)
{
  BtpTree* tree = (BtpTree*) data;

  g_return_val_if_fail (tree->new_parent, FALSE);
  g_return_val_if_fail (tree->switch_timeout != 0, FALSE);

  BTPP (1, "switch_timeout_cb\n");
  BTPP (7, "SWITCH TIMEOUT %s:%d\n", tree->new_parent->hostname, tree->new_parent->port);

  /* Mark the node as a bad parent */
  tree->new_parent->is_bad_parent = TRUE;

  /* Remove any switches that we might be about to send */
  b_conn_remove_packets (tree->new_parent->conn, B_PACKET_TYPE_SWITCH);

  /* Reset the new parent search */
  tree->new_parent = 0;
  tree->switch_timeout = 0;

  /* Request neighbors and start again */
  schedule_request_neighbors (tree, TRUE);

  return FALSE;
}

