/* BTP library - Banana Tree Protocol
 * Copyright (C) 1999-2001  The Regents of the University of Michigan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */


#include "btp_node.h"
#include "btp_debug.h"


/* **************************************** */



BtpTree*
btp_tree_new (BPeer* peer, GURL* url)
{
  BtpTree* tree;

  tree = g_new0 (BtpTree, 1);
  tree->peer = peer;
  tree->url = url;
  tree->ref_count = 1;

  return tree;

}




/*

  Delete the BTP tree.  Nodes may still be around at the end of this
  call, but we the only callbacks we should get are write callbacks.

  Last checked:	???

 */
void
btp_tree_delete (BtpTree* tree)
{
  GSList* i;

  g_return_if_fail (tree);
  BTPP (1, "btp_tree_delete %p\n", tree);

  /* Delete the children and siblings lists */
  g_slist_free (tree->children);
  g_slist_free (tree->siblings);
  tree->children = NULL;
  tree->siblings = NULL;

  /* Delete all nodes.  Note that the nodes may not be deleted until
     their out queues are empty.  Send packet needs 'me' to determine
     the port, so we delete it last. */
  for (i = tree->nodes; i != NULL; )
    {
      BtpNode* node = (BtpNode*) i->data;
      i = i->next;	/* node might be deleted */

      g_return_if_fail(node);

      if (node != tree->me)
	btp_node_delete(node);
    }
  btp_node_delete(tree->me);

  g_slist_free(tree->nodes);	
  tree->nodes = NULL;

  if (tree->switch_timeout)
    g_assert(g_source_remove(tree->switch_timeout));

  if (tree->request_neighbors_timeout)
    g_assert(g_source_remove(tree->request_neighbors_timeout));

  if (tree->find_new_parent_timeout)
    g_assert(g_source_remove(tree->find_new_parent_timeout));

  /* Zero myself out to be safe */
  memset(tree, 0, sizeof(*tree));

  g_free(tree);
}



/* **************************************** */

void
btp_tree_print (FILE* file, BtpTree* tree)
{
  gchar* url_str;
  GSList* i;

  g_return_if_fail (tree);

  fprintf (file, "\nBtpTree\n");
  url_str = gnet_url_get_nice_string(tree->url);
  fprintf (file, "url:        %s\n", url_str);
  g_free (url_str);
  fprintf (file, "passive:    %d\n", tree->passive);
  fprintf (file, "ref_count:  %d\n", tree->ref_count);
  fprintf (file, "rcto:       %d\n", tree->request_neighbors_timeout != 0);
  fprintf (file, "fnpto:      %d\n", tree->find_new_parent_timeout != 0);
  fprintf (file, "swto:       %d\n", tree->switch_timeout != 0);
  fprintf (file, "\n");

  
  fprintf (file, "me:\n");
  btp_node_print (file, tree->me);
  fprintf (file, "root:\n");
  btp_node_print (file, tree->root);
  fprintf (file, "parent:\n");
  btp_node_print (file, tree->parent);
  fprintf (file, "gparent:\n");
  btp_node_print (file, tree->gparent);
  fprintf (file, "new_parent:\n");
  btp_node_print (file, tree->new_parent);
  fprintf (file, "\n");

  fprintf (file, "Nodes: \n");
  for (i = tree->nodes; i != NULL; i = i->next)
    btp_node_print (file, (BtpNode*) i->data);
  fprintf (file, "\n");

  fprintf (file, "Children: \n");
  for (i = tree->children; i != NULL; i = i->next)
    btp_node_print (file, (BtpNode*) i->data);

  fprintf (file, "Sibling: \n");
  for (i = tree->siblings; i != NULL; i = i->next)
    btp_node_print (file, (BtpNode*) i->data);


}


/* **************************************** */


gboolean
btp_tree_is_me (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return (node->tree->me == node);
}


gboolean
btp_tree_is_root (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return (node->tree->root == node);
}


gboolean
btp_tree_is_parent (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return (node->tree->parent == node);
}


gboolean
btp_tree_is_gparent (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return (node->tree->gparent == node);
}


gboolean
btp_tree_is_new_parent (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return (node->tree->new_parent == node);
}


gboolean
btp_tree_is_related (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return btp_tree_is_root (node) 	||
    btp_tree_is_parent (node) 		||
    btp_tree_is_new_parent (node) 	||
    btp_tree_is_child (node) 		||
    btp_tree_is_sibling (node);

}


gboolean
btp_tree_is_sibling (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return (g_slist_find(node->tree->siblings, node) != NULL);
}



void
btp_tree_add_sibling (BtpNode* node)
{
  g_return_if_fail (node);
  g_return_if_fail (node->tree);
  g_return_if_fail (node != node->tree->me);

  if (!g_slist_find(node->tree->siblings, node))
    node->tree->siblings = g_slist_prepend (node->tree->siblings, node);
}


void
btp_tree_remove_sibling (BtpNode* node)
{
  g_return_if_fail (node);
  g_return_if_fail (node->tree);

  node->tree->siblings = g_slist_remove (node->tree->siblings, node);
}


gboolean
btp_tree_is_child (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return (g_slist_find(node->tree->children, node) != NULL);
}


void	
btp_tree_add_child (BtpNode* node)
{
  g_return_if_fail (node);
  g_return_if_fail (node->tree);
  g_return_if_fail (node != node->tree->me);

  if (!g_slist_find(node->tree->children, node))
    node->tree->children = g_slist_prepend (node->tree->children, node);
}


void
btp_tree_remove_child (BtpNode* node)
{
  g_return_if_fail (node);
  g_return_if_fail (node->tree);

  node->tree->children = g_slist_remove (node->tree->children, node);
}

guint
btp_tree_degree (BtpTree* tree)
{
  guint degree = 0;

  g_return_val_if_fail (tree, 0);

  if (tree->parent)
    degree++;

  degree += g_slist_length (tree->children);

  return degree;
}


guint
btp_tree_parent_degree (BtpTree* tree)
{
  g_return_val_if_fail (tree, 0);

  if (!tree->parent)
    return 0;

  return 1 /* parent */ + 1 /* me */ + g_slist_length (tree->siblings);
}


gboolean
btp_tree_is_node (BtpNode* node)
{
  g_return_val_if_fail (node, FALSE);
  g_return_val_if_fail (node->tree, FALSE);

  return (g_slist_find(node->tree->nodes, node) != NULL);
}


void	
btp_tree_add_node (BtpNode* node)
{
  g_return_if_fail (node);
  g_return_if_fail (node->tree);

  if (!g_slist_find(node->tree->nodes, node))
    node->tree->nodes = g_slist_prepend (node->tree->nodes, node);
}


void
btp_tree_remove_node (BtpNode* node)
{
  g_return_if_fail (node);
  g_return_if_fail (node->tree);

  node->tree->nodes = g_slist_remove (node->tree->nodes, node);
}


BtpNode* 
btp_tree_find_node (BtpTree* tree, gchar* hostname, gint port)
{
  GSList* i;

  g_return_val_if_fail (tree, NULL);
  g_return_val_if_fail (hostname, NULL);


  for (i = tree->nodes; i != NULL; i = i->next)
    {
      BtpNode* node = (BtpNode*) i->data;

      g_return_val_if_fail (node, NULL);
      g_return_val_if_fail (node->hostname, NULL);

      if (port == node->port &&
	  strcmp(hostname, node->hostname) == 0)
	return node;
    }

  return NULL;
}


/* ******************** */


void
btp_tree_send_mcast (BtpTree* tree, BtpNode* not_node, const void* buffer, guint16 length)
{
  GSList* ci;

  g_return_if_fail (tree);
  g_return_if_fail (buffer);

  /* Send to parent (if I have a parent) */
  if (tree->parent && tree->parent != not_node)
    {
      BPacket* packet = b_packet_new_mcast (tree->parent->conn, buffer, length);
      b_conn_send_packet (tree->parent->conn, packet);
    }

  /* Send to children */
  for (ci = tree->children; ci != NULL; ci = ci->next)
    {
      BtpNode* child_node = (BtpNode*) ci->data;

      if (child_node != not_node)
	{
	  BPacket* packet = b_packet_new_mcast (child_node->conn, buffer, length);
	  b_conn_send_packet (child_node->conn, packet);
	}
    }
}


void
btp_tree_send_mcast_children (BtpTree* tree, const void* buffer, guint16 length)
{
  g_return_if_fail (tree);
  g_return_if_fail (buffer);

  btp_tree_send_mcast (tree, tree->parent, buffer, length);
}



void
btp_tree_send_ucast (BtpTree* tree, BtpNode* node, const void* buffer, guint16 length)
{
  BPacket* packet;

  g_return_if_fail (tree);
  g_return_if_fail (node);
  g_return_if_fail (buffer);

  packet = b_packet_new_ucast (node->conn, buffer, length);
  b_conn_send_packet (node->conn, packet);
}


