/* Jungle Monkey
 * Copyright (C) 1999-2001  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "ggui.h"
#include "ggui_channels.h"
#include "ggui_chat.h"
#include "ggui_config.h"
#include "ggui_download.h"
#include "ggui_files.h"
#include "ggui_preferences.h"
#include "ggui_properties.h"
#include "ggui_search.h"
#include "ggui_debug.h"
#include "jmutil.h"
#include "pixmaps.h"

#include "util/util.h"
#include "util/jmgnet.h"
#include "util/poll.h"
#include "crp/crp.h"
#include "btp/b_conn.h"
#include "jmmsearch/jmmsp_debug.h"
#include "jmchat/jmchat_debug.h"
#include "jmchannel/jmchannel_debug.h"
#include "mtp/mtp_debug.h"

#include <signal.h>


static void menubar_enable (void);
static void menubar_disable (void);

static void crp_func (const gchar* hostname, gint port, CRPStatus status, 
		      GInetAddr* addr, gpointer user_data);

static void show_file (FILE* file, gchar* title);

gboolean on_jmw_destroy (GtkWidget* widget, gpointer user_data);
static void sig_int_cb (int ignore);
static void sig_pipe_cb (int ignore);
static gboolean final_destroy_cb(gpointer user_data);
static void conn_zero_cb (void);

void on_jmw_notebook_switch_page (GtkNotebook     *notebook,
				  GtkNotebookPage *page,
				  guint            page_num);




/* ******************** */
/* GUI			*/

GtkWidget* 	jmw;
GtkToolbar*     jmw_toolbar;
GladeXML* 	jmw_xml = NULL;
GtkWidget*	jmw_notebook;


/* **************************************** */


static gboolean shutting_down = FALSE;

static gchar*	crp_server = "crp://junglemonkey.net";
static CRPID*   crp_id;

static gboolean disable_restore = FALSE;


int
main(int argc, char* argv[])
{
  const gchar* url = NULL;
  JMUTIL_POPT_VARS;

  struct poptOption jm_options[] = 
  {
#ifndef JM_ENABLE_GNOME
    {"help", 'h', POPT_ARG_NONE, NULL, 1, 
     "Show this help message", NULL},
    {"usage", '\0', POPT_ARG_NONE, NULL, 2, 
     "Display brief usage message", NULL},
    {"version", 'V', POPT_ARG_NONE, NULL, 3,
     "Output verson information and exit", NULL},
#endif


    JMUTIL_POPT_ARGS,

    {"crp-server", '\0', POPT_ARG_STRING, &crp_server, 0,
     "URL of CRP server (default is crp://junglemonkey.net)", "URL"},

    {"disable-restore", '\0', POPT_ARG_NONE, &disable_restore, 0,
     "Disable restoring channels and files", NULL},

    {"home", '\0', POPT_ARG_STRING, &jm_home, 0, 
     "Set the JM home directory (default is ~/jm)", "HOME"},

    {NULL, '\0', 0, NULL, 0, NULL, NULL}
  };

  poptContext ctx;

  ggui_debug_flags = 0xffff;
  jmchannel_debug_flags = 0xffff;
  jmchat_debug_flags = 0xffff;
  jmmsp_debug_flags = 0xffff;
/*    mtp_debug_flags = 0xffff; */


  /* ******************** */
  /* Library init	  */

#ifdef ENABLE_NLS
  {
    gtk_set_locale();
    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);
  }
#endif


#ifdef JM_ENABLE_GNOME
  {

    gnome_init_with_popt_table ("Jungle Monkey", VERSION, 
				argc, argv,
				jm_options, 0, &ctx);
    glade_gnome_init();
  }
#else
  {
    int rc;

    gtk_init(&argc, &argv);
    glade_init();

    /* ******************** */
    /* Parse command line	  */
    ctx = poptGetContext(NULL, argc, argv, jm_options, 0);
    poptSetOtherOptionHelp(ctx, _("[OPTION]... [URL]"));

    while ((rc = poptGetNextOpt(ctx)) > 0)
      {
	switch (rc)
	  {
	  case 1:	  goto help;	  break;
	  case 2:	  goto usage;	  break;
	  case 3:   	  goto version;	  break;
	  }
      }
  }
#endif

  /* Get the URL */
  url = poptGetArg(ctx);
  poptFreeContext(ctx);

  if (!url)
    url = JM_DEFAULT_CHANNEL;


  /* ******************** */
  /* Initialize GUI	  */

  /* Workaround strict poll */
# ifdef HAVE_STRICT_POLL
    g_main_set_poll_func (gnu_poll);
# endif

  /* Load the XML */
# ifdef JM_ENABLE_GNOME
    jmw_xml = ggui_get_glade_xml("jmw", "jm-gnome.glade");
# else
    jmw_xml = ggui_get_glade_xml("jmw", "jm.glade");
# endif
  g_assert (jmw_xml);
  glade_xml_signal_autoconnect (jmw_xml);	

  jmw = glade_xml_get_widget (jmw_xml, "jmw");
  jmw_notebook = glade_xml_get_widget (jmw_xml, "jmw_notebook");
  jmw_toolbar = GTK_TOOLBAR(glade_xml_get_widget (jmw_xml, "jm_toolbar"));

  g_assert (jmw);
  g_assert (jmw_notebook);
  g_assert (jmw_toolbar);

  /* Load config file */
  ggui_config_init ();

  /* Load pixmaps */
  ggui_load_pixmaps ();


  /* **************************************** */
  /* Warn if I'm root 			      */

  if (getuid() == 0)
    {
      ggui_show_message ("warning", 
			 _("DO NOT RUN JUNGLE MONKEY AS ROOT! \n"
			   "If you do, you're asking for trouble."));
    }


  /* Set the Jungle Monkey home directory.  This is set in this order:
     
     1) command line
     2) Gnome config (if Gnome support)
     3) HOME + ~/


     Try to get it from the JMHOME environment variable.  If that
     isn't set, use HOME/jm.

  */
  if (jm_home == NULL)
    {
      jm_home = ggui_config_get_string("/Jungle Monkey/General/home", NULL);
      if (jm_home == NULL)
	{
	  jm_home = g_get_home_dir(); 	/* don't delete */
	  g_assert (jm_home != NULL);

	  jm_home = g_strconcat(jm_home, G_DIR_SEPARATOR_S, "jm", NULL);
	}
    }

  /* Strip off the trailing / if necesssary */
  if (jm_home && *jm_home && jm_home[strlen(jm_home)-1] == G_DIR_SEPARATOR)
    jm_home[strlen(jm_home)-1] = '\0';

  /* Create the JM home directory if it doesn't exist. */
  if (!mkdirr(jm_home, 0700))
    {
      ggui_show_message ("warning", 
			 _("Cannot create Jungle Monkey home directory \"%s\""), 
			 jm_home);
    }

  /* Create a .jm directory for configuration files */
  jm_config_home = g_strconcat (g_get_home_dir(), G_DIR_SEPARATOR_S, ".jm", NULL);
  if (!mkdirr (jm_config_home, 0700))
    {
      ggui_show_message ("warning", 
			 _("Cannot create Jungle Monkey configuration home directory \"%s\""), 
			 jm_config_home);
    }


  /* **************************************** */
  /* Initialize the backend 		      */

  if (JMUTIL_INIT)
    {
      /* Disable GUI */
      ggui_disable ();
    }
  else
    {
      /* Enable GUI */
      ggui_enable ();

      /* Check the configuration */
      ggui_check_network_config ();

    }

  /* Catch sig-int and sig-pipe */
  signal (SIGINT, sig_int_cb);
  signal (SIGPIPE, sig_pipe_cb);

  gtk_main();

  return 0;

#ifndef JM_ENABLE_GNOME
 usage:
  poptPrintUsage(ctx, stdout, 0);
  exit(EXIT_SUCCESS);

 help:
  g_print (_("`jm' - Jungle Monkey, a distributed file sharing program\n"));
  poptPrintHelp(ctx, stdout, 0);
  g_print (_("Report bugs to <jm-dev@junglemonkey.net>\n"));
  exit(EXIT_SUCCESS);

 version:
  g_print (_("jm %s\n"
	     "Written by David A. Helder\n"
	     "\n"
	     "Copyright (C) 1999-2001 The Regents of the University of Michigan\n"
	     "This is free software; see the source for copying conditions.  There is NO\n"
	     "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
	   VERSION);
  exit(EXIT_SUCCESS);
#endif
}




/* ******************** */

void
ggui_check_network_config (void)
{
  GURL* crp_url;

  /* Check the BTP hostname and port.  We won't bother checking MTP
     for now. */

  g_return_if_fail (jm_hostname);
  g_return_if_fail (jm_btp_iface);
  g_return_if_fail (crp_server);

  crp_url = gnet_url_new (crp_server);
  g_return_if_fail (crp_url);

  crp_id = crp_check (crp_url, jm_hostname, jm_btp_bpeer->port, 
		      crp_func, NULL);

  gnet_url_delete (crp_url);
}



static void 
crp_func (const gchar* hostname, gint port, CRPStatus status, 
	  /* callee owns */ GInetAddr* addr, gpointer user_data)
{
  gchar* addr_str = NULL;

  crp_id = NULL;
  g_return_if_fail (jm_mtp_server);

  if (addr)
    addr_str = gnet_inetaddr_get_canonical_name (addr);

  switch (status)
    {
    case CRP_STATUS_BAD_PAIR:
      ggui_show_message ("warning", 
			 _("JM is configured to use %s as your address \n"
			   "and %d as the port for incoming connections. \n"
			   "However, the CRP server thinks your hostname is \n"
			   "%s and cannot connect to port %d. \n"
			   "If you do not fix this, other JM users will not \n"
			   "be able to connect to you. If you are behind a \n"
			   "firewall or NAT, you need to configure it to allow \n"
			   "connections to pass through to this port and \n"
			   "port %d. Please reconfigure JM or contact your \n"
			   "network administrator."),
			 hostname, port, addr_str, port, 
			 jm_mtp_server->server->port);
      break;

    case CRP_STATUS_BAD_NAME:
      ggui_show_message ("warning", 
			 _("JM is configured to use %s as your address. \n"
			   "However, the CRP server thinks your address is \n"
			   "%s. If you do not fix this, other JM users \n"
			   "will not be able to connect to you. Please \n"
			   "reconfigure and restart JM."),
			 hostname, addr_str);
      break;

    case CRP_STATUS_BAD_PORT:
      ggui_show_message ("warning", 
			 _("JM is configured to use %d as the port for \n"
			   "incoming connections. However, the CRP server \n"
			   "cannot connect to this port. If you do not fix \n"
			   "this, other JM users will not be able to connect \n"
			   "to you. If you are behind a firewall or NAT, you \n"
			   "need to configure it to allow connections to pass \n"
			   "through to this port and port %d. Contact your \n"
			   "network administrator if you do not know how to \n"
			   "do this."),
			 port, jm_mtp_server->server->port);
      break;


    case CRP_STATUS_GOOD:
    case CRP_STATUS_ERROR:
      ;
    }

  gnet_inetaddr_delete (addr);
  g_free (addr_str);
}






/* ******************** */


void
ggui_enable (void)
{
  static gboolean enabled = FALSE;

  if (enabled)
    return;
  enabled = TRUE;

  g_return_if_fail (jmw_notebook);
  
  ggui_channel_init ();
  ggui_file_init ();
  ggui_properties_init ();
  ggui_preferences_init ();
  ggui_chat_init ();
  ggui_search_init ();
  ggui_download_init ();

  ggui_channel_load();
  ggui_file_load();
  ggui_chat_load();
  ggui_search_load();

  menubar_enable ();

  ggui_file_show_toolbar ();

  gtk_widget_set_sensitive (jmw_notebook, TRUE);

}


void
ggui_disable (void)
{
  g_return_if_fail (jmw_notebook);
  
  menubar_disable ();
  gtk_widget_set_sensitive (jmw_notebook, FALSE);
}


static void
menubar_enable (void)
{
  GtkWidget* open_channel;
  GtkWidget* create_root_channel;

  open_channel = glade_xml_get_widget(jmw_xml, "open_channel_mi");
  create_root_channel = glade_xml_get_widget(jmw_xml, "create_root_channel_mi");

  g_return_if_fail (open_channel != NULL);
  g_return_if_fail (create_root_channel != NULL);

  gtk_widget_show(open_channel);
  gtk_widget_show(create_root_channel);
}


static void
menubar_disable (void)
{
  GtkWidget* open_channel;
  GtkWidget* create_root_channel;;

  open_channel = glade_xml_get_widget(jmw_xml, "open_channel_mi");
  create_root_channel = glade_xml_get_widget(jmw_xml, "create_root_channel_mi");

  g_return_if_fail (open_channel != NULL);
  g_return_if_fail (create_root_channel != NULL);

  gtk_widget_hide(open_channel);
  gtk_widget_hide(create_root_channel);
}


void
ggui_save_hashtable (const gchar* name, GHashTable* hash_table, GHFunc hfunc)
{
  gchar* path;
  FILE*  file;

  if (disable_restore)
    return;

  /* Save created chats */
  path = g_strconcat (jm_config_home, G_DIR_SEPARATOR_S, name, NULL);
  file = fopen (path, "w");
  if (file)
    {
      g_hash_table_foreach (hash_table, hfunc, file);
      fclose (file);
    }
  g_free (path);
}


void
ggui_load_elf (const gchar* name, void (*func)(const ElfNode*))
{
  gchar* path;
  GList* list;
  GList* i;

  if (disable_restore)
    return;

  path = g_strconcat (jm_config_home, G_DIR_SEPARATOR_S, name, NULL);
  list = elf_read_list_path (path);
  for (i = list; i != NULL; i = i->next)
    func ((ElfNode*) i->data);
  elf_delete_list (list);
  g_free (path);
}



/* ************************************************************ */
/* GUI utility functions					*/

void
ggui_activate (const GURL* url)
{
  g_return_if_fail (url);

  if (gnet_url_is_scheme (url, "jm"))
    ggui_channel_join (url);

  else if (gnet_url_is_scheme (url, "mtp"))
    {
      if (ggui_file_have (url))
	/* Do nothing */;
      else
	ggui_download_get (url, NULL, NULL);
    }

  else if (gnet_url_is_scheme (url, "jmchat"))
    ggui_chat_join(url);

  else if (gnet_url_is_scheme (url, "jmmsp"))
    ggui_search_join (url);

  else
    {
#     ifdef JM_ENABLE_GNOME
      {
	gchar* url_str = gnet_url_get_nice_string (url);
	if (url_str)
	  {
	    gnome_url_show (url_str);
	    g_free (url_str);
	  }
      }
#     endif
    }

  return;
}


void
ggui_get_pixmaps (const GURL* url, GdkPixmap** pixmapp, GdkBitmap** maskp)
{
  GdkPixmap* pixmap 	= NULL;
  GdkBitmap* mask	= NULL;

  g_return_if_fail (pixmapp);
  g_return_if_fail (maskp);
  g_return_if_fail (url);


  if (gnet_url_is_scheme (url, "jm"))
    {
      if (ggui_channel_is_up (url))
	{
	  pixmap = channel_pixmap;
	  mask   = channel_mask;
	}
      else if (ggui_channel_is_down (url))
	{
	  pixmap = down_channel_pixmap;
	  mask   = down_channel_mask;
	}
      else
	{
	  pixmap = gray_channel_pixmap;
	  mask   = gray_channel_mask;
	}
    }

  else if (gnet_url_is_scheme (url, "mtp"))
    {
      gint percent;

      if (ggui_file_have (url) ||
	  (percent = ggui_download_get_percent_complete (url)) == 100)
	{
	  pixmap = page_pixmap;
	  mask   = page_mask;
	}
      else if (percent == -2)
	{
	  pixmap = down_page_pixmap;
	  mask   = down_page_mask;
	}
      else
	{
	  pixmap = gray_page_pixmap;
	  mask   = gray_page_mask;
	}
    }

  else if (gnet_url_is_scheme (url, "jmchat"))
    {
      if (ggui_chat_is_up (url))
	{
	  pixmap = chat_pixmap;
	  mask   = chat_mask;
	}
      else if (ggui_chat_is_down (url))
	{
	  pixmap = down_chat_pixmap;
	  mask   = down_chat_mask;
	}
      else
	{
	  pixmap = gray_chat_pixmap;
	  mask   = gray_chat_mask;
	}
    }
  
  else if (gnet_url_is_scheme (url, "jmmsp"))
    {
      if (ggui_search_is_up (url))
	{
	  pixmap = search_pixmap;
	  mask   = search_mask;
	}
      else if (ggui_search_is_down (url))
	{
	  pixmap = down_search_pixmap;
	  mask   = down_search_mask;
	}
      else
	{
	  pixmap = gray_search_pixmap;
	  mask   = gray_search_mask;
	}
    }

  else
    {
      pixmap = url_pixmap;
      mask   = url_mask;
    }

  if (pixmapp)	*pixmapp = pixmap;
  if (maskp)    *maskp   = mask;
}
		 

/* **************************************** */


/*

  Load <glade_filename>, which should have widget <name> in it.  If we
  can't load the file, we fail.

 */
GladeXML* 
ggui_get_glade_xml(gchar* name, gchar* glade_filename)
{
  GladeXML* gxml = NULL;
  gchar* jm_glade;

  /* First check local directory */
  jm_glade = g_strconcat(".", G_DIR_SEPARATOR_S, 
			 "glade", G_DIR_SEPARATOR_S, 
			 glade_filename, NULL);
  if (file_exists(jm_glade))
    {
      gxml = glade_xml_new(jm_glade, name);
    }
  g_free  (jm_glade);

  if (gxml == NULL)
    {
      jm_glade = g_strconcat(JM_DATADIR, G_DIR_SEPARATOR_S, 
			     "glade", G_DIR_SEPARATOR_S, 
			     glade_filename, NULL);

      if (file_exists(jm_glade))
	gxml = glade_xml_new(jm_glade, name);

      g_free (jm_glade);
    }

  if (gxml == NULL)
    {
      ggui_show_message ("error", _("Could not load glade file %s. "
				    "We looked in the directories ./glade and %s/glade. "
				    "Maybe JM is not correctly installed or libglade is "
				    "out-of-date."), glade_filename, JM_DATADIR);
    }

  return gxml;
}




#ifndef JM_ENABLE_GNOME

static void
on_messagebox_ok (gpointer button, gpointer dialog)
{
  gtk_widget_destroy (GTK_WIDGET(dialog));
}

#endif



void 
ggui_show_message(const gchar* type, const gchar* format, ...)
{
  va_list args;
  gchar* message;

  g_return_if_fail (type != NULL);
  g_return_if_fail (format != NULL);

  va_start (args, format);
  message = g_strdup_vprintf(format, args);
  va_end (args);

  g_return_if_fail (message != NULL);

#ifdef JM_ENABLE_GNOME
  {
    GtkWidget* message_box;

    message_box = gnome_message_box_new (message, type, GNOME_STOCK_BUTTON_OK, NULL);
    if (jmw)
      gnome_dialog_set_parent (GNOME_DIALOG(message_box), GTK_WINDOW(jmw));
    gtk_widget_show(message_box);
  }
#else
  {
    GtkWidget* dialog;
    GtkWidget* label;
    GtkWidget* ok;

    dialog = gtk_dialog_new();
    GTK_WINDOW(dialog)->position = GTK_WIN_POS_MOUSE;

    label = gtk_label_new (message);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, FALSE, FALSE, 4);
    gtk_widget_show(label);

    ok = gtk_button_new_with_label (_("OK"));
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), ok, FALSE, FALSE, 4);
    gtk_signal_connect (GTK_OBJECT (ok), "clicked",
			GTK_SIGNAL_FUNC(on_messagebox_ok), (gpointer) dialog);  
    gtk_widget_show (ok);


    if (!SAFESTRCMP(type, "info"))
      {
	gtk_window_set_title(GTK_WINDOW(dialog), _("INFORMATION"));
      }
    else if (!SAFESTRCMP(type, "warning"))
      {
	gtk_window_set_title(GTK_WINDOW(dialog), _("WARNING"));
      }
    else if (!SAFESTRCMP(type, "error"))
      {
	gtk_window_set_title(GTK_WINDOW(dialog), _("ERROR"));
      }
    else
      {
	gtk_window_set_title(GTK_WINDOW(dialog), _("Message"));
      }

    gtk_widget_show(dialog);

  }
#endif

  g_free (message);
}


void
ggui_set_tooltip(GtkWidget* widget, const gchar* tip)
{
  GtkTooltips* tooltip;

  tooltip = gtk_tooltips_new();

  gtk_tooltips_set_tip(tooltip, widget, tip, NULL);
}




void
ggui_clist_update_row2 (GtkCList* clist, gint row,
			gchar* text1, gchar* text2,
			GdkPixmap* pixmap, GdkBitmap* mask)
{
  gchar*     old_text = NULL;		/* don't delete */
  GdkPixmap* old_pixmap = NULL;
  GdkBitmap* old_mask = NULL;

  g_return_if_fail (clist != NULL);
  g_return_if_fail (row >= 0);

  gtk_clist_get_pixtext (clist, row, 0, &old_text, NULL,
			 &old_pixmap, &old_mask);

  if ((!old_text && text1) ||
      (old_text && !text1) || 
      (old_text && text1 && SAFESTRCMP(old_text, text1)) ||
      (old_pixmap != pixmap)  ||
      (old_mask   != mask))
    {
      gtk_clist_set_pixtext (clist, row, 0, text1, 5,
			     pixmap, mask);
    }


  old_text = NULL;
  gtk_clist_get_text (clist, row, 1, &old_text);

  if ((!old_text && text2) || 
      (old_text && !text2) || 
      (old_text && text2 && SAFESTRCMP(old_text, text2)))
    {
      gtk_clist_set_text (clist, row, 1, text2);
    }
}





/* ************************************************************ */
/* GUI signal code						*/


gboolean
on_jmw_destroy (GtkWidget* ignore, gpointer user_data)
{
  if (shutting_down)
    return TRUE;

  shutting_down = TRUE;

  /* Kill CRP */
  if (crp_id)
    crp_cancel (crp_id);
  crp_id = NULL;

  /* Save state */
  ggui_channel_save ();
  ggui_file_save ();
  ggui_chat_save ();
  ggui_search_save ();

  /* Shutdown GUI's */
  ggui_channel_shutdown ();
  ggui_file_shutdown ();
  ggui_chat_shutdown ();
  ggui_download_shutdown ();
  ggui_search_shutdown (); /* after file */
  ggui_preferences_shutdown ();
  ggui_properties_shutdown ();

  /* Save jmhome */
  ggui_config_set_string ("/Jungle Monkey/General/home", jm_home);
      
  if (jm_btp_bpeer)
    ggui_config_set_int ("/Jungle Monkey/General/BTP last port", jm_btp_bpeer->port);
  if (jm_mtp_server)
    ggui_config_set_int ("/Jungle Monkey/General/MTP last port", jm_mtp_server->server->port);

  ggui_config_shutdown ();

  /* Close the window */
  g_assert (jmw != NULL);

  gtk_widget_destroy(jmw);

  b_conn_num_conns_zero = conn_zero_cb;
  fprintf (stderr, "btp conns = %d\n", b_conn_num_conns);
  if (!jm_running || jmutil_shutdown())
    {
      gtk_main_quit();
      return TRUE;
    }
  else
    {
      g_timeout_add (5000, final_destroy_cb, NULL);
    }

  return TRUE;	/* we handled the signal */
}



static void
sig_int_cb (int ignore)
{
  /* Reset signal handler */
  signal (SIGINT, SIG_DFL);

  /* Ignore if shutting down */
  if (shutting_down)
    return;

  /* Just call on_jmw_destory */
  on_jmw_destroy (NULL, NULL);
}


static void
sig_pipe_cb (int ignore)
{
  /* Reset both signal handlers */
  signal (SIGPIPE, SIG_DFL);
  signal (SIGINT, SIG_DFL);

  g_print (_("\nERROR: Caught SIGPIPE signal.  This is a bug.  Dumping core.\n"));
  abort ();
}


static gboolean
final_destroy_cb(gpointer user_data)
{
  gtk_main_quit();
  return FALSE;
}


static void
conn_zero_cb (void)
{
  fprintf (stderr, "conn_zero_cb\n");
  gtk_main_quit();
}



/* **************************************** */

void
on_jmw_notebook_switch_page (GtkNotebook     *notebook,
			     GtkNotebookPage *page,
			     guint            page_num)
{
  GtkContainer* cont;
  GList* kids;
  GList* i;

  /* Sometimes this is called when we shutdown */
  if (shutting_down)
    return;

  /* Clear the toolbar */
  cont = GTK_CONTAINER (jmw_toolbar);
  kids = gtk_container_children (cont);
  for (i = kids; i != NULL; i = i->next)
    gtk_container_remove (cont, GTK_WIDGET(i->data));
  g_list_free (kids);
  
  /* Call appropriate function */
  switch (page_num)
    {
    case 0: ggui_file_show_toolbar (); 		break;
    case 1: ggui_download_show_toolbar (); 	break;
    case 2: ggui_chat_show_toolbar (); 		break;
    case 3: ggui_search_show_toolbar (); 	break;
    default: g_assert_not_reached ();
    }
}


void
on_open_channel(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
/*    ggui_menu_open_channel(); FIX */
}

void
on_create_root_channel(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
/*    ggui_menu_create_root_channel(); FIX */
}


# ifdef JM_ENABLE_GNOME

void
on_mtp_server_info (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  FILE* file = NULL;

  if (!jm_mtp_server)
    {
      ggui_show_message ("warning", _("MTP Server not started"));
      return;
    }

  /* Create temp file */
  file = tmpfile();
  g_return_if_fail (file);

  /* Print MTP server info and rewind */
  mtp_server_print (jm_mtp_server, file);
  show_file (file, _("MTP Server Information"));
  fclose (file);
}


void
on_btp_peer_info (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  FILE* file = NULL;

  if (!jm_btp_bpeer)
    {
      ggui_show_message ("warning", _("BTP Peer not started"));
      return;
    }

  /* Create temp file */
  file = tmpfile();
  g_return_if_fail (file);

  /* Print BTP server info and rewind */
  btp_print (file, jm_btp_bpeer);
  show_file (file, _("BTP Peer Information"));
  fclose (file);
}


static void
show_file (FILE* file, gchar* title)
{
  GtkWidget* dialog;
  GtkWidget* gl;

  g_return_if_fail (file);
  g_return_if_fail (title);

  rewind (file);

  /* Create dialog */
  dialog = gnome_dialog_new (title, GNOME_STOCK_BUTTON_OK, NULL);
  GTK_WINDOW(dialog)->position = GTK_WIN_POS_MOUSE;
  gnome_dialog_set_default (GNOME_DIALOG(dialog), 0);
  gtk_signal_connect (GTK_OBJECT(dialog), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), NULL);

  /* Create Gnome less */
  gl = gnome_less_new();
  gnome_less_show_filestream (GNOME_LESS(gl), file);
  gtk_box_pack_start (GTK_BOX(GNOME_DIALOG(dialog)->vbox), gl, TRUE, TRUE, 0);
  gtk_widget_show (gl);

  gtk_widget_show (dialog);
}



# endif /* JM_ENABLE_GNOME */



void
on_jm_homepage (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
# ifdef JM_ENABLE_GNOME
    {
      gnome_url_show ("http://www.junglemonkey.net");
    }
# endif
}


void
on_jm_manual (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
# ifdef JM_ENABLE_GNOME
    {
      gchar *tmp;
      tmp = gnome_help_file_find_file ("jm", "jm.html");
      if (tmp) 
	{
	  gnome_help_goto (0, tmp);
	  g_free (tmp);
	}
    }
# endif
}


void
on_jm_faq (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
# ifdef JM_ENABLE_GNOME
    {
      gchar *tmp;
      tmp = gnome_help_file_find_file ("jm", "faq.html");
      if (tmp) 
	{
	  gnome_help_goto (0, tmp);
	  g_free (tmp);
	}
    }
# endif
}


void
on_about(gpointer callback_data, guint callback_action, GtkWidget* widget)
{

#ifdef JM_ENABLE_GNOME
  GtkWidget* about;
  const gchar* authors[] = {_("David A. Helder"), NULL};

  about = gnome_about_new (_("Jungle Monkey"),
			   VERSION,
			   _("Copyright (C) 1999-2001 The Regents of the University of Michigan"),
			   authors,
			   NULL,		/* Comment */
			   NULL 		/* Logo	*/
			   );
  gnome_dialog_set_parent (GNOME_DIALOG(about), GTK_WINDOW(jmw));
  gtk_widget_show(about);
#endif
}
