/* Jungle Monkey
 * Copyright (C) 1999-2001  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "jmchannel.h"
#include "jmchannel_debug.h"
#include "jmrindex.h"

#include "util/elf.h"

#include <string.h>
#include <stdlib.h>

static JMChannel* channel_new (BPeer* bpeer, Btp* btp);

static void 	send_remove_r (JMChannel* channel, JMAnn* ann);

static void     create_index_and_update (JMChannel* channel, JMAnn* ann);
static void	remove_r (JMChannel* channel, JMAnn* ann);

static JMAnn*	get_or_create (JMChannel* channel, const gchar* path, gboolean dirs_only);

static void 	reset (JMChannel* channel);
static void	reset_r (JMAnn* ann);
static void     delete_rindex_hfunc (gpointer key, gpointer value, gpointer user_data);

static void     set_send_updates_timer (JMChannel* channel);
static void     send_updates (JMChannel* channel);
static void	delay_send_updates (JMChannel* channel, guint delay);
static gboolean send_updates_cb (gpointer data);
static void 	send_join (JMChannel* channel);
static void	send_update (JMChannel* channel, JMAnn* ann);
static void	send_remove (JMChannel* channel, JMAnn* ann);

static gboolean gc_cb (gpointer data);
static void     gc_r (JMChannel* channel, JMAnn* ann, Timeval* timeofday);
static void     remove_old_ann (JMChannel* channel, JMAnn* ann);

static gboolean packet_cb (Btp* btp, const void* buf, guint len, gpointer user_data);
static void     error_cb (Btp* btp, gpointer user_data);


static JMChannel*
channel_new (BPeer* bpeer, Btp* btp)
{
  JMChannel* channel;
  JMAnn*     root_ann;

  /* Create a new JMChannel */
  channel = g_new0 (JMChannel, 1);
  channel->peer = bpeer;
  channel->btp  = btp;
  channel->path_to_ann   = g_hash_table_new (g_str_hash, g_str_equal);
  channel->ann_to_rindex = g_hash_table_new (NULL, NULL);
  channel->ann_to_index  = g_hash_table_new (NULL, NULL);

  /* Create root announcment */
  root_ann = jmann_new ("");
  root_ann->path = g_strdup("/");
  Gettimeofday (&root_ann->last_update);
  root_ann->elf = elf_new (NULL);
  jmann_set_attr (root_ann, "name", root_ann->path);
  channel->root_ann = root_ann;
  g_hash_table_insert (channel->path_to_ann, root_ann->path, root_ann);

  /* Initialize BTP */
  if (btp)
    {
      btp->packet_func 	    = packet_cb;
      btp->error_func  	    = error_cb;
      btp->packet_user_data = channel;
      btp->error_user_data  = channel;

      /* Send JOIN */
      send_join (channel);

      /* Schedule UPDATE */
      set_send_updates_timer (channel);

      /* Schedule garbage collection */
      channel->gc_timer = g_timeout_add (JMCHANNEL_TIMEOUT_TIME, gc_cb, channel);
    }

  return channel;
}


JMChannel* 
jmchannel_create (BPeer* bpeer, const gchar* name)
{
  JMChannel* channel;
  Btp*       btp;

  g_return_val_if_fail (name, NULL);

  /* Create a new BTP session */
  btp = btp_create (bpeer, name);
  if (btp == NULL)
    return NULL;

  /* Create a new JMChannel */
  channel = channel_new (bpeer, btp);
  g_return_val_if_fail (channel, NULL);
  channel->is_local = TRUE;

  /* Set up URL */
  channel->url = btp_get_url (btp);
  gnet_url_set_protocol (channel->url, "jm");

  /* Localize root announcement */
  channel->root_ann->is_local = TRUE;
  channel->local_anns = g_list_prepend (channel->local_anns, channel->root_ann);
  channel->num_local_anns++;
  channel->send_anns = g_list_prepend (channel->send_anns, channel->root_ann);
  jmann_set_attr_url (channel->root_ann, "url", channel->url);

  /* Send announcement */
  send_update (channel, channel->root_ann);

  return channel;
}


JMChannel* 
jmchannel_join (BPeer* bpeer, const GURL* url)
{
  JMChannel* channel;
  Btp* btp;

  g_return_val_if_fail (bpeer, NULL);
  g_return_val_if_fail (url, NULL);

  /* Join the BTP */
  btp = btp_join (bpeer, url);
  if (!btp)
    return NULL;

  /* Create a new JMChannel */
  channel = channel_new (bpeer, btp);
  g_return_val_if_fail (channel, NULL);

  /* Save URL */
  channel->url  = gnet_url_clone (url);

  return channel;
}


JMChannel* 
jmchannel_setup (BPeer* bpeer, const GURL* url)
{
  JMChannel* channel;

  g_return_val_if_fail (bpeer, NULL);
  g_return_val_if_fail (url, NULL);

  /* Create a new JMChannel */
  channel = channel_new (bpeer, NULL);
  g_return_val_if_fail (channel, NULL);

  /* Save URL */
  channel->url  = gnet_url_clone (url);

  return channel;
}


gboolean
jmchannel_rejoin (JMChannel* channel)
{
  Btp* btp;

  g_return_val_if_fail (channel, TRUE);
  g_return_val_if_fail (!channel->btp, TRUE);

  /* Join the BTP */
  btp = btp_join (channel->peer, channel->url);
  if (!btp)
    return TRUE;

  channel->btp = btp;

  /* Initialize BTP */
  btp->packet_func 	= packet_cb;
  btp->error_func  	= error_cb;
  btp->packet_user_data = channel;
  btp->error_user_data  = channel;

  /* We aren't down any more */
  channel->is_down = FALSE;

  /* Send JOIN */
  send_join (channel);

  /* Send updates */
  send_updates (channel);

  /* Schedule garbage collection if we didn't already */
  if (!channel->gc_timer)
    {
      channel->gc_timer = 
	 g_timeout_add (JMCHANNEL_TIMEOUT_TIME, gc_cb, channel);
    }

  return FALSE;
}


void
jmchannel_leave (JMChannel* channel)
{
  g_return_if_fail (channel);

  /* Close the BTP */
  if (channel->btp)
    {
      /* Send removes for all local files. */
      send_remove_r (channel, channel->root_ann);

      /* Reset the channel */
      reset (channel);

      /* Recreate hash tables */
      channel->path_to_ann = g_hash_table_new (g_str_hash, g_str_equal);
      channel->ann_to_rindex = g_hash_table_new (NULL, NULL);
      channel->ann_to_index = g_hash_table_new (NULL, NULL);

      /* Add root back */
      g_hash_table_insert (channel->path_to_ann, channel->root_ann->path, 
			   channel->root_ann);
      if (channel->root_ann->is_local)
	{
	  channel->local_anns = 
	    g_list_prepend (channel->local_anns, channel->root_ann);
	  channel->num_local_anns++;
	  channel->send_anns = 
	    g_list_prepend (channel->send_anns, channel->root_ann);
	}
    }
}


static void
send_remove_r (JMChannel* channel, JMAnn* ann)
{
  g_return_if_fail (channel);
  g_return_if_fail (ann);

  /* Send removes for local anns.  Removes are recursive, so we don't
     need to send removes for children of locals. */

  if (ann->is_local)
    send_remove (channel, ann);

  else
    {
      GList* i;

      for (i = ann->kids; i != NULL; i = i->next)
	send_remove_r (channel, (JMAnn*) i->data);
    }
}


void
jmchannel_delete (JMChannel* channel)
{
  if (!channel)
    return;

  g_free (channel->url);
  reset (channel);
  reset_r (channel->root_ann);

  memset (channel, 0, sizeof(*channel));
  g_free (channel);
}


void
jmchannel_set_mtp (JMChannel* channel, 
		   GURL* mtp_url, MtpServer* mtp_server)
{
  g_return_if_fail (channel);
  g_return_if_fail ((mtp_url && mtp_server) || (!mtp_url && !mtp_server));

  if (channel->mtp_url)
    {
      gnet_url_delete (channel->mtp_url);
      channel->mtp_url = NULL;
    }
  
  channel->mtp_url = gnet_url_clone (mtp_url);
  channel->mtp_server = mtp_server;
}


JMAnn*
jmchannel_get (JMChannel* channel, const gchar* path)
{
  g_return_val_if_fail (channel,NULL);
  g_return_val_if_fail (path,   NULL);

  return (JMAnn*) g_hash_table_lookup (channel->path_to_ann, path);
}


JMAnn*	   
jmchannel_add_elf (JMChannel* channel, JMAnn* parent, 
		   const gchar* name, const GURL* url, 
		   const ElfNode* elf)
{
  gchar*   path;
  JMAnn*   ann;

  g_return_val_if_fail (channel, NULL);
  g_return_val_if_fail (parent,  NULL);
  g_return_val_if_fail (name,    NULL);

  /* Make sure doesn't exist */
  path = jmann_make_path (parent, name);
  if (jmchannel_has (channel, path))
    {
      g_free (path);
      return NULL;
    }

  /* Create ann */
  ann = jmann_new (name);
  if (url)
    ann->url = gnet_url_clone (url);
  ann->path = path;
  ann->is_local = TRUE;
  if (elf)
    ann->elf = elf_clone (elf);
  else
    ann->elf = elf_new (NULL);

  /* Set name and URL attr */
  jmann_set_attr (ann, "name", path);
  if (url)
    jmann_set_attr_url (ann, "url", url);

  /* Add to lists */
  g_hash_table_insert (channel->path_to_ann, path, ann);
  channel->local_anns = g_list_prepend (channel->local_anns, ann);
  channel->num_local_anns++;
  ann->parent = parent;
  parent->kids = g_list_prepend (parent->kids, ann);


  /* Send announcement and add to send_anns or an index */

  /* If this is an index file, do nothing.  We get announced by the
     obj that triggered the index files creation.  We do not go into
     an index file. */
  if (jmann_is_index (ann))
    {
      /* Do nothing. */
      JMCP (2, "jmchannel_add: index, do nothing\n");
    }

  /* Create an index if there are enough files, or the current index
     is full. */
  else if (channel->mtp_url && channel->mtp_server &&
	   ((!channel->cur_index && 
	     channel->num_local_anns >= JMCHANNEL_MIN_INDEX_SIZE) ||
	    (channel->cur_index && 
	     channel->cur_index->num_anns >= JMCHANNEL_MAX_INDEX_SIZE)))
    {
      JMCP (2, "jmchannel_add: create index\n");
      create_index_and_update (channel, ann);
    }

  /* Add to the index file if there is one.  This is the common case.
     We add some delay to the update for better aggregation.  The
     append will cause an unmirror.  Note we do not send an update. */
  else if (channel->cur_index)
    {
      JMCP (2, "jmchannel_add: add to index\n");
      
      jmindex_add (channel->cur_index, ann);
      delay_send_updates (channel, JMCHANNEL_UPDATE_DELAY);
      /* (this will cause an update to be sent soon) */
    }

  /* Otherwise, send an announcement now.  There isn't enough for an
     index file yet (and I'm not the index file). */
  else
    {
      JMCP (2, "jmchannel_add: send announcement\n");

      channel->send_anns = g_list_prepend (channel->send_anns, ann);
      send_update (channel, ann);
    }

  /* Make sure we're sending updates */
  if (!channel->update_timer)
    set_send_updates_timer (channel);

  /* Upcall */
  if (channel->info_func)
    channel->info_func (channel, JMCHANNEL_UPDATE, ann, channel->user_data);

  return ann;
}


JMAnn*	   
jmchannel_add_path_elf (JMChannel* channel, const gchar* path, 
			const GURL* url, const ElfNode* elf)
{
  JMAnn* ann;
  JMAnn* parent;
  const gchar* name;
  gchar* name_unesc;

  g_return_val_if_fail (channel,      NULL);
  g_return_val_if_fail (path, 	      NULL);
  g_return_val_if_fail (*path == '/', NULL);

  /* Do quick lookup (which should fail) */
  ann = jmchannel_get (channel, path);
  g_return_val_if_fail (!ann, NULL);

  /* Get parent */
  parent = get_or_create (channel, path, TRUE);
  g_return_val_if_fail (parent, NULL);

  /* Figure out name */
  for (name = path; *name; ++name) ; /* find end */
  --name;
  while (name >= path && (*name != '/' || name[-1] == '\\')) --name; /* find start */
  name++;
  name_unesc = strunesc (name, "/\\", '\\');

  ann = jmchannel_add_elf (channel, parent, name_unesc, url, elf);
  g_free (name_unesc);

  return ann;
}



static void
create_index_and_update (JMChannel* channel, JMAnn* ann)
{
  gchar    name[] = ".jmXXXXXX.jmi";
  gint	   j;
  GURL*    url;
  gchar*   path;
  ElfNode* elf;
  JMIndex* index;
  GList*   i;
  JMAnn*   iann;

  /* Create a name */
  do
    {
      for (j = 3; j <= 8; ++j)
	name[j] = (rand() % ('z' - 'a')) + 'a';
    } 
  while (jmann_get_child (channel->root_ann, name));

  /* Create index URL */
  url = gnet_url_clone (channel->mtp_url);
  path = g_strconcat (jmchannel_get_name (channel), "/", name, NULL);
  gnet_url_set_resource (url, path);
  g_free (path);

  /* Create index elf */
  elf = elf_new_format (NULL, "b", "index", TRUE);

  /* Create the index */
  index = jmindex_new (channel);
  g_return_if_fail (index);

  /* Add index announcement */
  iann = jmchannel_add_elf (channel, channel->root_ann, name, url, elf);
  gnet_url_delete (url);
  elf_delete (elf);
  g_return_if_fail (iann);

  /* Initialize */
  channel->send_anns = g_list_prepend (channel->send_anns, iann);
  channel->cur_index = index;
  channel->indexes = g_list_prepend (channel->indexes, index);
  g_hash_table_insert (channel->ann_to_index, iann, index);
  index->me_ann = iann;

  /* Move all non-index send_anns to the index */
  for (i = channel->send_anns; i != NULL; )
    {
      JMAnn* sann = (JMAnn*) i->data;
      i = i->next;
      
      /* Skip indexes */
      if (jmann_is_index (sann))
	continue;

      /* Remove from list */
      channel->send_anns = g_list_remove (channel->send_anns, sann);

      /* Add to index */
      jmindex_add (index, sann);
    }

  /* Rewrite the index */
  jmindex_add (index, ann);
  jmindex_rewrite (index);
  jmann_set_attr_sha (iann, "sha",    index->sha);
  jmann_set_attr_int (iann, "length", index->length);

  /* Send update for this announcement */
  send_update (channel, ann);

  /* Delay before sending update for index */
  delay_send_updates (channel, JMCHANNEL_UPDATE_DELAY);
}


void
jmchannel_remove (JMChannel* channel, JMAnn* ann)
{
  JMAnn* parent;
  GList* i;

  g_return_if_fail (channel);
  g_return_if_fail (ann);
  g_return_if_fail (ann->is_local);
  g_return_if_fail (ann->parent);
  g_return_if_fail (!jmann_is_index(ann));

  /* Send announcement */
  send_remove (channel, ann);

  /* Delete the announcement and any children */
  parent = ann->parent;
  parent->kids = g_list_remove (parent->kids, ann);
  remove_r (channel, ann);

  /* Update upcall to parent */
  if (parent && channel->info_func)
    channel->info_func (channel, JMCHANNEL_UPDATE, 
			parent, channel->user_data);

  /* Find it's index and remove it.  TODO: Hash this. */
  for (i = channel->indexes; i != NULL; i = i->next)
    {
      JMIndex* index = (JMIndex*) i->data;

      if (jmindex_have(index, ann))
	{
	  jmindex_remove (index, ann);
	  delay_send_updates (channel, JMCHANNEL_UPDATE_DELAY);
	  break;
	}
    }
}


static void
remove_r (JMChannel* channel, JMAnn* ann)
{
  GList* i;
  JMRIndex* rindex;

  for (i = ann->kids; i != NULL; i = i->next)
    remove_r (channel, (JMAnn*) i->data);
  g_list_free (ann->kids);

  if (ann->is_local)
    {
      channel->local_anns = g_list_remove (channel->local_anns, ann);
      channel->num_local_anns--;
      channel->send_anns  = g_list_remove (channel->send_anns, ann);
    }

  if (ann == channel->root_ann)
    channel->root_ann = NULL;
    
  g_hash_table_remove (channel->path_to_ann, ann->path);

  /* Delete rindex */
  rindex = g_hash_table_lookup (channel->ann_to_rindex, ann);
  if (rindex)
    {
      g_hash_table_remove (channel->ann_to_rindex, ann);
      jmrindex_delete (rindex);
    }

  /* Remove upcall */
  if (channel->info_func)
    channel->info_func (channel, JMCHANNEL_REMOVE, ann, channel->user_data);

  jmann_delete (ann);
}



void
jmchannel_recv (JMChannel* channel, const void* buf, guint len)
{
  GList* nodes;
  GList* i;

/*    fprintf (stderr, "IN\n"); */
/*    fwrite (buf, len, 1, stderr); */

  nodes = elf_read_list (buf, len);
  for (i = nodes; i != NULL; i = i->next)
    {
      ElfNode* elf = (ElfNode*) i->data;

      if (!strcmp(elf->name, "join"))
	{
	  /* Send an update as soon as possible (if we have
             announcements to send) */
	  if (channel->send_anns)
	    send_updates (channel);
	}

      else if (!strcmp(elf->name, "update"))
	{
	  gchar*    name;
	  JMAnn*    ann;
	  gchar*    url_str;
	  GURL*	    url;
	  gboolean  is_index;
	  JMRIndex* rindex;
	  
	  name = elf_get_attribute (elf, "name");
	  if (!name) continue;

	  /* Get/create the announcement */
	  ann = get_or_create (channel, name, FALSE);
	  if (!ann) continue;

	  JMCP (4, "jmchannel_recv update %s\n", name);

	  /* Warn if it's local */
	  if (ann->is_local)
	    {
	      g_warning ("Received announcement for local file %s on %s\n", 
			 ann->path, jmchannel_get_name(channel));
	      continue;
	    }

	  /* Get URL */
	  url_str = elf_get_attribute (elf, "url");
	  if (url_str)
	    url = gnet_url_new (url_str);
	  else
	    url = NULL;

	  /* Check URL */
	  if (!url && ann->url)
	    g_warning ("URL missing in %s\n", ann->path);
	  else if (url && ann->url && !gnet_url_equal(url, ann->url))
	    g_warning ("URL changed in %s\n", ann->path);
	  else if (url && !ann->url)
	    {
	      ann->url = url;
	      url = NULL;
	    }
	  /* otherwise, urls are equal or no urls at all */

	  gnet_url_delete (url);

	  /* Save elf */
	  elf_delete (ann->elf);
	  ann->elf = elf_clone (elf);

	  /* Process index */
	  is_index = jmann_get_attr_bool (ann, "index");
	  rindex = (JMRIndex*) g_hash_table_lookup (channel->ann_to_rindex, ann);
	  if (rindex || (is_index && ann->url))
	    {
	      if (!rindex)
		{
		  rindex = jmrindex_new (channel, ann);
		  g_hash_table_insert (channel->ann_to_rindex, ann, rindex);
		}

	      jmrindex_update (rindex);
	    }

	  /* Update upcall */
	  if (channel->info_func)
	    channel->info_func (channel, JMCHANNEL_UPDATE, ann, channel->user_data);
	}

      else if (!strcmp(elf->name, "remove"))
	{
	  gchar* name;
	  JMAnn* ann;
	  JMAnn* parent;

	  /* Get name and announcement.  Ignore if we haven't seen
	     the announcement */
	  name = elf_get_attribute (elf, "name");
	  if (!name) continue;
	  ann = jmchannel_get (channel, name);
	  if (!ann)  continue;

	  /* Ignore if for root ann */
	  if (ann == channel->root_ann)
	    continue;

	  /* Remove the announcement and any children */
	  parent = ann->parent;
	  remove_r (channel, ann);

	  /* Update upcall to parent */
	  if (parent)
	    {
	      parent->kids = g_list_remove (parent->kids, ann);
	      if (channel->info_func)
		channel->info_func (channel, JMCHANNEL_UPDATE, 
				    parent, channel->user_data);
	    }
	}

      else
	g_warning ("Received weird announcement \"%s\"\n", elf->name);

    }


  elf_delete_list (nodes);
}


static JMAnn*	
get_or_create (JMChannel* channel, const gchar* path, gboolean dirs_only)
{
  JMAnn*  parent;
  gint    num_tokens;
  gchar** tokens;
  const gchar*  p;
  gint    i;

  g_return_val_if_fail (channel,      NULL);
  g_return_val_if_fail (path,  	      NULL);
  g_return_val_if_fail (*path == '/', NULL);

  /* Do quick lookup */
  parent = jmchannel_get (channel, path);
  if (parent)
    {
      /* Note last update */
      Gettimeofday (&parent->last_update);

      return parent;
    }

  /* Tokenize */
  p = path + 1;
  num_tokens = strchrcount (p, "/") + 1;  /* overestimate tokens */
  tokens = g_new0(gchar*, num_tokens + 1);
  i = 0;
  while (*p)
    {
      const gchar* q;
      gchar* tkn;

      /* Find end of first token */
      for (q = p; *q && (*q != '/' || q[-1] == '\\'); ++q) /* nothing */;

      /* Allocate space */
      tokens[i++] = tkn = g_malloc (q - p + 1);

      /* Copy over */
      for (q = p; *q && (*q != '/' || q[-1] == '\\'); ++q)
	{
	  /* Skip \ in \\ or \/ */
	  if (*q == '\\' && (q[1] == '/' || q[1] == '\\'))
	    ++q;
	  *tkn = *q;
	  ++tkn;
	}
      *tkn = '\0';
      if (!*q) break;

      p = q + 1;
    }

  /* Start with root as parent */
  parent = channel->root_ann;
  g_return_val_if_fail (parent, NULL);

  for (i = 0; tokens[i]; ++i)
    {
      JMAnn* ann;

      /* Return if second-to-last dir and dirs_only */
      if (dirs_only && !tokens[i+1])
	return parent;

      /* Get the child */
      ann = jmann_get_child (parent, tokens[i]);
      if (!ann)
	{
	  ann = jmann_new (tokens[i]);
	  ann->path = jmann_make_path (parent, tokens[i]);
	  ann->elf = elf_new (NULL);
	  ann->parent = parent;
	  parent->kids = g_list_prepend (parent->kids, ann);

	  g_hash_table_insert (channel->path_to_ann, ann->path, ann);

	  /* Set local if dir */
	  if (dirs_only && tokens[i+1])
	    ann->is_local = TRUE;

	  /* Update (if not last token) */
	  if (tokens[i+1] && channel->info_func)
	    channel->info_func (channel, JMCHANNEL_UPDATE, 
				ann, channel->user_data);
	}

      /* Note last update */
      Gettimeofday (&ann->last_update);

      /* Advance parent */
      parent = ann;
    }

  g_strfreev (tokens);

  return parent;
}


/* **************************************** */

static void
reset (JMChannel* channel)
{
  GList* i;

  g_return_if_fail (channel);

  /* Delete the BTP */
  if (channel->btp)
    {
      btp_leave (channel->btp);
      channel->btp = NULL;
    }

  /* Stop updates and garbage collection */
  if (channel->update_timer)
    {
      g_source_remove (channel->update_timer);
      channel->update_timer = 0;
    }

  if (channel->gc_timer)    
    {
      g_source_remove (channel->gc_timer);
      channel->gc_timer = 0;
    }

  /* Clear rindex hashtable */
  g_hash_table_foreach (channel->ann_to_rindex, delete_rindex_hfunc, NULL);
  g_hash_table_destroy (channel->ann_to_rindex);
  channel->ann_to_rindex = NULL;

  /* Delete all announcements (but root) */
  /* NOTE: Root announcement is not destroyed */
  for (i = channel->root_ann->kids; i != NULL; i = i->next)
    reset_r ((JMAnn*) i->data);
  g_list_free (channel->root_ann->kids);
  channel->root_ann->kids = NULL;

  /* Clear ann hashtable */
  g_hash_table_destroy (channel->path_to_ann);
  channel->path_to_ann = NULL;

  /* Clear local announcements */
  g_list_free (channel->local_anns);
  channel->local_anns      = NULL;
  channel->num_local_anns = 0;
  g_list_free (channel->send_anns);
  channel->send_anns      = NULL;

  /* Delete indexes */
  channel->cur_index = NULL;
  for (i = channel->indexes; i != NULL; i = i->next)
    jmindex_delete ((JMIndex*) i->data);
  g_list_free (channel->indexes);
  channel->indexes = NULL;
  g_hash_table_destroy (channel->ann_to_index);
  channel->ann_to_index = NULL;
}


static void
reset_r (JMAnn* ann)
{
  GList* i;

  for (i = ann->kids; i != NULL; i = i->next)
    reset_r ((JMAnn*) i->data);
  g_list_free (ann->kids);

  jmann_delete (ann);
}


static void
delete_rindex_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  jmrindex_delete ((JMRIndex*) value);
}


/* ******************** */

static void
set_send_updates_timer (JMChannel* channel)
{
  g_return_if_fail (channel);
  g_return_if_fail (!channel->update_timer);

  channel->update_timer = 	
    g_timeout_add (JMCHANNEL_UPDATE_TIME, send_updates_cb, channel);
}


/* 

   Send an update.  We keep track of when the last update was sent.
   If we sent an update recently, we will delay sending the update.
   This prevents an attacker from flooding the group with updates by
   sending a series of joins (smurf-attack-like).

 */
static void
send_updates (JMChannel* channel)
{
  Timeval wantsend;
  Timeval maxsend;

  g_return_if_fail (channel);

  /* Cancel any update timer */
  rand_timer_cancel (&channel->update_timer);

  /* Get next possible send time */
  Gettimeofday (&wantsend);
  maxsend = channel->update_tv;
  timeraddmsec (&maxsend, JMCHANNEL_UPDATE_TIME_MAX);

  /* Check if we can send an update now */
  if (timerleq (&maxsend, &wantsend)) /* maxsend <= wantsend */
    {
      JMCP (2, "send_updates now\n");
      send_updates_cb (channel);
    }

  /* Otherwise, send an update real soon */
  else
    {
      guint time;

      timersub (&maxsend, &wantsend, &wantsend);
      time = wantsend.tv_sec * 1000 + wantsend.tv_usec / 1000;

      rand_timer_set (&channel->update_timer, time, 0,
		      send_updates_cb, channel);
      JMCP (2, "send_updates in %d ms\n", time);
    }
}


static void
delay_send_updates (JMChannel* channel, guint delay)
{
  g_return_if_fail (channel);

  /* Cancel any update timer */
  rand_timer_cancel (&channel->update_timer);

  /* Reschedule */
  channel->update_timer = 	
    g_timeout_add (delay, send_updates_cb, channel);
}


static gboolean
send_updates_cb (gpointer data)
{
  JMChannel* channel = (JMChannel*) data;
  GList* i;
  gchar buf[JMCHANNEL_MAX_PACKET];
  guint off = 0;

  g_return_val_if_fail (channel, FALSE);

  /* Reset update timer */
  channel->update_timer = 0;

  JMCP (2, "send_updates_cb\n");

  /* If BTP is gone, then stop sending updates */
  if (!channel->btp)
    return FALSE;

  /* Note update time */
  Gettimeofday (&channel->update_tv);

  /* Send all updates */
  for (i = channel->send_anns; i != NULL; i = i->next)
    {
      JMAnn* ann = (JMAnn*) i->data;
      gchar* tbuf;
      guint  tlen;

      JMCP (2, "send_updates_cb: send %s:%s\n", 
	    jmchannel_get_name(channel), ann->path);

      /* Rewrite if it's a dirty index file */
      if (jmann_is_index(ann))
	{
	  JMIndex* index;

	  index = g_hash_table_lookup(channel->ann_to_index, ann);
	  g_return_val_if_fail (index, FALSE);
	  jmindex_rewrite (index);
	  jmann_set_attr_sha (ann, "sha", index->sha);
	  jmann_set_attr_int (ann, "length", index->length);
	}

      /* Serialize */
      elf_set_name (ann->elf, "update");
      elf_write (ann->elf, &tbuf, &tlen);

      /* Check if too big */
      if (tlen > JMCHANNEL_MAX_PACKET || tlen < 0)
	{
	  g_warning ("Announcement for %s bad size: %d\n", 
		     ann->name, tlen);
	  g_free (tbuf);
	  continue;
	}

      /* Check if send current buffer */
      if (off + tlen > JMCHANNEL_MAX_PACKET)
	{
/*  	  fprintf (stderr, "OUT\n"); */
/*  	  fwrite (buf, off, 1, stderr); */

	  btp_send (channel->btp, buf, off);
	  off = 0;
	}

      /* Copy the packet into the buffer */
      memcpy (&buf[off], tbuf, tlen);
      off += tlen;

      g_free (tbuf);
    }

  /* Send any data left */
  if (off)
    {
/*  	  fprintf (stderr, "OUT2\n"); */
/*  	  fwrite (buf, off, 1, stderr); */
	  btp_send (channel->btp, buf, off);
    }

  /* Update timer */
  set_send_updates_timer (channel);

  return FALSE;
}


static void
send_join (JMChannel* channel)
{
  gchar* buf;
  guint  len;

  g_return_if_fail (channel);

  JMCP (2, "send_join\n");

  if (!channel->btp)
    return;

  elf_write_format (&buf, &len, "join", "");
  btp_send (channel->btp, buf, len);
  g_free (buf);
}


static void
send_update (JMChannel* channel, JMAnn* ann)
{
  gchar* buf;
  guint  len;

  g_return_if_fail (channel);
  g_return_if_fail (ann);
  g_return_if_fail (ann->elf);

  JMCP (2, "send_update %s:%s\n", jmchannel_get_name(channel), ann->path);

  if (!channel->btp)
    return;

  elf_set_name (ann->elf, "update");
  elf_write (ann->elf, &buf, &len);
  btp_send (channel->btp, buf, len);
  g_free (buf);
}


static void
send_remove (JMChannel* channel, JMAnn* ann)
{
  gchar* buf;
  guint  len;

  g_return_if_fail (channel);
  g_return_if_fail (ann);

  JMCP (2, "send_remove %s:%s\n", jmchannel_get_name(channel), ann->path);

  if (!channel->btp)
    return;

  elf_write_format (&buf, &len, "remove", "s", "name", ann->path);
  btp_send (channel->btp, buf, len);
  g_free (buf);
}




/* ******************** */

static gboolean
gc_cb (gpointer data)
{
  JMChannel* channel = (JMChannel*) data;
  Timeval timeofday;

  g_return_val_if_fail (channel, FALSE);

  JMCP (2, "gc_cb\n");

  /* Get time */
  Gettimeofday (&timeofday);

  /* Recursively garbage collect */
  gc_r (channel, channel->root_ann, &timeofday);

  return TRUE;
}


static void
gc_r (JMChannel* channel, JMAnn* ann, Timeval* timeofday)
{
  GList*  i;
  Timeval diff;

  /* Check all children */
  for (i = ann->kids; i != NULL; )
    {
      JMAnn* kid = (JMAnn*) i->data;
      i = i->next;
      gc_r (channel, kid, timeofday);
    }

  /* Ignore if root, local, or still have children */
  if (!ann->parent || ann->is_local || ann->kids)
    return;

  /* Remove if timed out */
  timersub(timeofday, &ann->last_update, &diff);
  if (diff.tv_sec > (JMCHANNEL_TIMEOUT_TIME / 1000))
    {
      JMCP (2, "gc_cb: remove %s\n", ann->path);

      /* Notify upper level */
      if (channel->info_func)
	channel->info_func (channel, JMCHANNEL_REMOVE_TIMEOUT, 
			    ann, channel->user_data);

      remove_old_ann (channel, ann);
    }

}


static void
remove_old_ann (JMChannel* channel, JMAnn* ann)
{
  JMRIndex* rindex;

  g_return_if_fail (!ann->is_local);
  g_return_if_fail (ann->parent);

  /* Remove from parent or top level */
  ann->parent->kids = g_list_remove (ann->parent->kids, ann);

  /* Remove from hashtable */
  g_hash_table_remove (channel->path_to_ann, ann->path);

  /* Delete rindex */
  rindex = (JMRIndex*) g_hash_table_lookup (channel->ann_to_rindex, ann);
  if (rindex)
    {
      jmrindex_delete (rindex);
      g_hash_table_remove (channel->ann_to_rindex, ann);
    }

  /* Delete ann */
  jmann_delete (ann);
}


/* ******************** */

static gboolean
packet_cb (Btp* btp, const void* buf, guint len, gpointer user_data)
{
  jmchannel_recv ((JMChannel*) user_data, buf, len);
  return FALSE;
}


static void
error_cb (Btp* btp, gpointer user_data)
{
  JMChannel* channel = (JMChannel*) user_data;

  /* Delete BTP */
  btp_leave (btp);
  channel->btp = NULL;

  /* Stop update timer */
  if (channel->update_timer)
    {
      g_source_remove (channel->update_timer);
      channel->update_timer = 0;
    }
  /* Keep garbage collection going. */

  /* Mark us as down */
  channel->is_down = TRUE;
  
  /* Error upcall */
  if (channel->error_func)
    channel->error_func (channel, channel->user_data);
}
