/* Jungle Monkey Multicast Search Protocol Client
 * Copyright (C) 2000-2001  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <signal.h>

#ifdef JM_ENABLE_GNOME
#include <popt-gnome.h>
#else
#include <popt.h>
#endif

#include "jmintl.h"
#include "util/util.h"

#include "jmchannel.h"
#include "jmchannel_debug.h"
#include "mtp/mtp_debug.h"



static void sig_int_cb (int ignore);

static void info_cb (JMChannel* channel, JMChannelType type, 
		     JMAnn* ann, gpointer user_data);
static void error_cb (JMChannel* channel, gpointer user_data);




/* ******************** */
/* Globals		*/

static int   		verbose = 0;

static BPeer*		bpeer;

static MtpServer*	mtp_server;
static GURL*		mtp_url;

static JMChannel*	jmchannel;


extern int 		btp_debug_level;
extern unsigned int 	btp_debug_flags;


int
main(int argc, char* argv[])
{
  const gchar* url = NULL;
  GMainLoop* main_loop;
  const gchar* name;

  struct poptOption jm_options[] = 
  {
    {"help", 'h', POPT_ARG_NONE, NULL, 1, 
     "Show this help message", NULL},
    {"usage", '\0', POPT_ARG_NONE, NULL, 2, 
     "Display brief usage message", NULL},
    {"version", 0, POPT_ARG_NONE, NULL, 3,
     "Output verson information and exit", NULL},
    {"verbose", 'v', POPT_ARG_NONE, &verbose, 0,
     "Print lots of stuff", NULL},

    {"debug-flags", 'd', POPT_ARG_INT, &jmchannel_debug_flags, 0, 
     "Set the MTP debug flags", "FLAGS"},

    {NULL, '\0', 0, NULL, 0, NULL, NULL}
  };

  poptContext ctx;
  int rc;
  int exit_status = EXIT_SUCCESS;


  /* ******************** */
  /* Initialize NLS 	  */
#ifdef ENABLE_NLS
  {
/*      gtk_set_locale(); */
    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);
  }
#endif

  /* ******************** */
  /* Parse command line	  */

  ctx = poptGetContext(NULL, argc, argv, jm_options, 0);
  poptSetOtherOptionHelp(ctx, _("[OPTION]... [URL] [FILES]..."));

  while ((rc = poptGetNextOpt(ctx)) > 0)
    {
      switch (rc)
	{
	case 1:	  goto help;	  break;
	case 2:	  goto usage;	  break;
	case 3:   goto version;	  break;
	}
    }

  url = poptGetArg(ctx);
  if (url == NULL)
    {
      g_print ("*** You must specify a URL.\n");
      goto usage;
    }

  /* Catch SIGINT */
  signal (SIGINT, sig_int_cb);

  /* ******************** */
  /* Create bpeer	  */

/*    btp_debug_level = 1; */
/*    btp_debug_flags = 0xFFFF; */
  jmchannel_debug_flags = 0xffff;
  mtp_debug_flags = 0xffff;

  bpeer = b_peer_new ("ask.eecs.umich.edu", NULL, FALSE);
  g_assert (bpeer);
  g_print ("BTP open at %s:%d\n", bpeer->hostname, bpeer->port);

  mtp_server = mtp_server_new ("ask.eecs.umich.edu", NULL, FALSE,
			       MTP_SERVER_TYPE_BOTH, FALSE);
  g_assert (mtp_server);
  g_print ("MTP open at %s:%d\n", mtp_server->hostname, 
	   mtp_server->server->port);

  mtp_url = gnet_url_new_fields ("mtp", mtp_server->hostname, 
				 mtp_server->server->port, NULL);
  g_assert (mtp_url);


  /* ******************** */
  /* Join/create jmmsp    */

  if (!strncmp(url, "jm://", sizeof("jm://")-1))		/* JOIN */
    {
      GURL* gurl;

      gurl = gnet_url_new (url);
      if (gurl == NULL)
	{
	  g_print ("*** Malformed URL: %s\n", url);
	  goto usage;
	}

      jmchannel = jmchannel_join (bpeer, gurl);
      if (!jmchannel)
	{
	  fprintf (stderr, "%s: Could not join channel %s\n", argv[0], url);
	  exit (EXIT_FAILURE);
	}

      g_print ("Joined %s\n", url);
    }
  else
    {
      jmchannel = jmchannel_create (bpeer, url);	/* CREATE */
      if (!jmchannel)
	{
	  fprintf (stderr, "%s: Could not join channel %s\n", argv[0], url);
	  exit (EXIT_FAILURE);
	}

      g_print ("Created jm://%s:%d/%s\n", 
	       bpeer->hostname, bpeer->port, url);

    }
  jmchannel->info_func  = info_cb;
  jmchannel->error_func = error_cb;
  jmchannel_set_mtp (jmchannel, mtp_url, mtp_server);

  while ((name = poptGetArg(ctx)) != NULL)
    {
      GURL* gurl = NULL;

      url = poptGetArg(ctx);
      if (url)
	{
	  gurl = gnet_url_new (url);
	  g_assert (gurl);
	}

      if (*name == '/')
	jmchannel_add_path (jmchannel, name, gurl);
      else if (jmchannel->root_ann)
	jmchannel_add (jmchannel, jmchannel->root_ann, name, gurl);
      else
	g_warning ("Bad name: %s\n", name);
    }

  poptFreeContext(ctx);


  /* ******************** */
  /* Start the main loop  */

  main_loop = g_main_new(FALSE);
  g_main_run(main_loop);

  exit(EXIT_SUCCESS);

 usage:
  poptPrintUsage(ctx, stdout, 0);
  exit(exit_status);

 help:
  g_print (_("`jmchannelclient' - Jungle Monkey Channel Client\n"));
  poptPrintHelp(ctx, stdout, 0);
  g_print (_("\n"
	     "Report bugs to <jm-dev@junglemonkey.net>\n"));
  exit(exit_status);

 version:
  g_print (_("jmmspclient %s\n"
	     "Written by David A. Helder\n"
	     "\n"
	     "Copyright (C) 1999-2001 The Regents of the University of Michigan\n"
	     "This is free software; see the source for copying conditions.  There is NO\n"
	     "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
	     VERSION);
  exit(exit_status);
}


/* **************************************** */


static void
sig_int_cb (int ignore)
{
  /* Reset signal handler */
  signal (SIGINT, SIG_DFL);

  /* Delete stuff */
  if (jmchannel)
    {
      jmchannel_leave (jmchannel);
      jmchannel_delete (jmchannel);
      jmchannel = NULL;
    }

  exit (EXIT_SUCCESS);
}


static void
info_cb (JMChannel* channel, JMChannelType type, 
	 JMAnn* ann, gpointer user_data)
{
  switch (type)
    {
    case JMCHANNEL_UPDATE: g_print ("UPDATE: "); break;
    case JMCHANNEL_REMOVE: g_print ("REMOVE: "); break;
    case JMCHANNEL_REMOVE_TIMEOUT: g_print ("REMOVE_TIMEOUT: "); break;
    default: g_assert_not_reached();
    }

  g_print ("%s:%s\n", jmchannel_get_name(channel), ann->path);
}


static void
error_cb (JMChannel* jmchannel, gpointer user_data)
{
  g_print ("JMChannel Error\n");
}
