/* Jungle Monkey
 * Copyright (C) 1999-2001  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "jmindex.h"
#include "jmchannel.h"
#include "jmann.h"
#include "jmchannel_debug.h"

#include <stdlib.h>



JMIndex*
jmindex_new (JMChannel* channel)
{
  char     pathname[] = "/tmp/jmXXXXXX";
  int      fd;
  JMIndex* index;

  g_return_val_if_fail (channel, NULL);

  /* Create temporary file */
  fd = mkstemp (pathname);
  g_return_val_if_fail (fd != -1, NULL);
  close (fd);

  /* Create index */
  index = g_new0 (JMIndex, 1);
  index->path     = g_strdup (pathname);
  index->is_dirty = TRUE;
  index->sha      = NULL;
  index->channel  = channel;

  return index;
}


void
jmindex_delete (JMIndex* index)
{
  if (!index)
    return;

  g_slist_free (index->anns);

  unlink (index->path);
  g_free (index->path);

  gnet_sha_delete (index->sha);

  if (index->mirror)
    mtp_unmirror (index->mirror);

  g_free (index);
}



/* ******************** */

void
jmindex_add (JMIndex* index, JMAnn* ann)
{
  g_return_if_fail (index);
  g_return_if_fail (ann);

  if (jmindex_have(index, ann))
    return;

  index->anns = g_slist_prepend (index->anns, ann);
  ++index->num_anns;
  jmindex_touch (index);
}


void
jmindex_remove (JMIndex* index, JMAnn* ann)
{
  g_return_if_fail (index);
  g_return_if_fail (ann);

  if (!jmindex_have(index, ann))
    return;

  index->anns = g_slist_remove (index->anns, ann);
  --index->num_anns;
  jmindex_touch (index);
}


void
jmindex_touch (JMIndex* index)
{
  g_return_if_fail (index);

  JMCP (8, "jmindex_touch: %s\n", index->path);

  /* Set dirty flag */
  index->is_dirty = TRUE;

  /* Unmirror */
  if (index->mirror)
    {
      mtp_unmirror (index->mirror);
      index->mirror = NULL;
    }
}


void
jmindex_rewrite (JMIndex* index)
{
  JMChannel* channel = index->channel;
  FILE*      file;
  GSList*    i;
  gint       rv;

  g_return_if_fail (index);

  JMCP (8, "jmindex_rewrite: %s\n", index->path);

  /* Unmirror */
  if (index->mirror)
    {
      mtp_unmirror (index->mirror);
      index->mirror = NULL;
    }

  /* Reset dirty flag */
  index->is_dirty = FALSE;

  /* Open the file */
  file = fopen (index->path, "w");
  g_return_if_fail (file);
  index->length = 0;

  /* Delete the old SHA and create new one */
  gnet_sha_delete (index->sha);
  index->sha = gnet_sha_new_incremental ();

  /* Write each announcement */
  for (i = index->anns; i != NULL; i = i->next)
    {
      JMAnn* ann = (JMAnn*) i->data;
      gchar* buf;
      guint  len;

      /* Create the announcement */
      elf_set_name (ann->elf, "update");
      elf_write (ann->elf, &buf, &len);
      g_return_if_fail (buf);

      /* Append it to the file */
      rv = fwrite (buf, len, 1, file);
      g_free (buf);
      g_return_if_fail (rv == 1);
      index->length += len;

      /* Update the SHA */
      gnet_sha_update (index->sha, buf, len);
    }

  /* Finalize new SHA */
  gnet_sha_final (index->sha);

  /* Close the file */
  fclose (file);

  /* Mirror again */
  if (channel->mtp_server)
    index->mirror = mtp_mirror_path (channel->mtp_server,
				     index->me_ann->url,
				     index->path);
}
