/* Jungle Monkey
 * Copyright (C) 1999-2001  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "jmchannel.h"
#include "jmchannel_debug.h"
#include "jmann.h"
#include "jmrindex.h"

#include <stdlib.h>
#include <errno.h>
#include <string.h>


static void download (JMRIndex* rindex);
static void process (JMRIndex* rindex);
static void file_func (MtpClient* client, MtpClientStatus status, gpointer user_data);


JMRIndex* 
jmrindex_new (JMChannel* channel, JMAnn* ann)
{
  JMRIndex* rindex;

  g_return_val_if_fail (channel,        NULL);
  g_return_val_if_fail (ann,            NULL);
  g_return_val_if_fail (!ann->is_local, NULL);

  rindex = g_new0 (JMRIndex, 1);
  rindex->channel = channel;
  rindex->ann     = ann;

  return rindex;
}


void
jmrindex_delete (JMRIndex* rindex)
{
  if (!rindex)
    return;

  if (rindex->client)
    mtp_client_delete (rindex->client);
  gnet_sha_delete (rindex->sha);

  if (rindex->have_file)
    unlink (rindex->path);
  g_free (rindex->path);

  g_free (rindex);
}



void
jmrindex_update (JMRIndex* rindex)
{
  GSHA* sha;

  /* Get SHA */
  sha = jmann_get_attr_sha (rindex->ann, "sha");

  /* Download index if we don't have it or the SHA changed. */
  if (!rindex->have_file ||
      (sha && rindex->sha && !gnet_sha_equal (rindex->sha, sha)))
    {
      download (rindex);
    }

  /* Otherwise, reprocess the index if we have it*/
  else if (rindex->have_file)
    {
      Timeval timeofday;
      Timeval next_process;

      /* Process if it's been a while since we processed it */
      Gettimeofday (&timeofday);
      next_process = rindex->last_process;
      timeraddmsec (&next_process, JMCHANNEL_UPDATE_TIME / 2);
      if (timerlt (&next_process, &timeofday))
	{
	  process (rindex);
	}
    }
}


static void
download (JMRIndex* rindex)
{
  MtpClient* client;

  g_return_if_fail (rindex);

  JMCP (0x10, "jmrindex_download: %s\n", rindex->ann->path);

  /* Delete old client */
  if (rindex->client)
    {
      mtp_client_delete (rindex->client);
      rindex->client = NULL;
    }

  /* Or, delete old file */
  else if (rindex->have_file)
    {
      unlink (rindex->path);
      rindex->have_file = FALSE;
    }

  /* Ignore if no URL */
  if (!rindex->ann->url)
    return;

  /* Create temporary file */
  if (!rindex->path)
    {
      char pathname[] = "/tmp/jmXXXXXX";
      int  fd;

      fd = mkstemp (pathname);
      g_return_if_fail (fd != -1);
      close (fd);
      rindex->path = g_strdup (pathname);
    }

  /* Save the SHA.  We save the SHA in index SHA so we can tell if the
     SHA changed while we were in the process of downloading it. */
  gnet_sha_delete (rindex->sha);
  rindex->sha = jmann_get_attr_sha (rindex->ann, "sha");

  /* Get the file */
  client = mtp_client_get_file (rindex->ann->url, rindex->path, 
				file_func, rindex);
  g_return_if_fail (client);
  rindex->client = client;
}



static void
process (JMRIndex* rindex)
{
  int rv;
  FILE* file = NULL;
  long  size;
  char* buf = NULL;
  static gboolean in_index = FALSE; 

  g_return_if_fail (rindex);
  g_return_if_fail (rindex->have_file);

  /* Don't process indexes in indexes */
  if (in_index)
    return;

  JMCP (0x10, "jmrindex_process %s\n", rindex->ann->path);

  /* Note time the index was processed */
  Gettimeofday (&rindex->last_process);

  /* Open the file */
  file = fopen (rindex->path, "r");
  if (!file) return;

  /* Get the size */
  size = FILE_size (file);
  if (size < 0) goto done;

  /* Allocate space for it */
  buf = g_malloc (size);
      
  /* Read the file */
  rv = fread (buf, size, 1, file);
  if (rv != 1)
    {
      g_warning ("Read error: %s\n", strerror(errno));
      goto done;
    }
  fclose (file); file = NULL;

  /* Process the index */
  in_index = TRUE;
  jmchannel_recv (rindex->channel, buf, size);
  in_index = FALSE;

 done:
  if (file) fclose (file);
  if (buf)  g_free (buf);

  /* TODO: Allow mirroring of index files */
}


static void
file_func (MtpClient* client, MtpClientStatus status, gpointer user_data)
{
  JMRIndex* rindex = (JMRIndex*) user_data;

  /* TODO: Incremental update on UPDATE? */

  JMCP (0x10, "jmrindex_file_func %s\n", rindex->ann->path);

  if (status == MTP_CLIENT_STATUS_OK)
    {
      GSHA* msha;

      msha = client->sha;	/* Steal SHA before delete */
      client->sha = NULL;

      mtp_client_delete (client);
      rindex->client = NULL;

      if (msha && rindex->sha && !gnet_sha_equal (msha, rindex->sha))
	{
	  gchar str1[2 * GNET_SHA_HASH_LENGTH + 1];
	  gchar str2[2 * GNET_SHA_HASH_LENGTH + 1];

	  gnet_sha_copy_string (msha, str1);
	  str1[2 * GNET_SHA_HASH_LENGTH] = '\0';
	  gnet_sha_copy_string (rindex->sha, str2);
	  str2[2 * GNET_SHA_HASH_LENGTH] = '\0';

	  g_warning ("Index had bad SHA: %s [%s] (is %s, should be %s)\n", 
		     rindex->ann->path, rindex->path, str1, str2);
	}

      rindex->have_file = TRUE;
      process (rindex);

      gnet_sha_delete (msha);
    }
  else if (status == MTP_CLIENT_STATUS_ERROR)
    {
      mtp_client_delete (client);
      rindex->client = NULL;
    }
}
